
///////////////////////////////////////////////////////////////
//
//	File:		sparse.c
//	Created:	20.12.2001
//	Purpose:	Sparse dynamic programming in sequence alignment
//
//	Comments:	
//	Authors:	Bogopolsky Gennady, EIMB (c) 2001
//
///////////////////////////////////////////////////////////////

#include "hier.h"
#include "windows.h"


kernel	Kernels_dop[MAX_KERNELS];
long	Dop_kernel_count;

//---------------------------------------------------
/*int kernel_weight_compare(const void * arg1, const void * arg2)
{
	const kernel* krn1 = (const kernel*)arg1;
	const kernel* krn2 = (const kernel*)arg2;
	if(krn1->score > krn2->score) return -1;
	else if (krn1->score < krn2->score) return 1;
	return 0;
}*/

//-----------------------------------------------------
int hier_init_pair_primer_families(sparse_alignment_hier_data* _data)
{
	register int	i, j;
	unsigned		idx;
	score_type		w, w1, w2;
	char			aa1, aa2, aax1, aax2;
	score_type		cutoff = _data->krn_params[0].pair_cutoff;
	long_pointer*	pair_families = (long_pointer*)_data->pair_families;

	assert(_data);
	memset(_data->pair_families, 0, sizeof(_data->pair_families));

	for(i = 0; i < 400; i++) {
		for(j = 0; j < 400; j++) {
			aa1 = i / 20;
			aa2 = i % 20;
			aax1 = j / 20;
			aax2 = j % 20;
			w1 = subst_matrix[aa1][aax1];
			if ( w1 >= 0 ) {
				w2 = subst_matrix[aa2][aax2];
				w = (w2 >= 0) ? w1+w2 : w1; /**/
			//w = subst_matrix[aa1][aax1] + subst_matrix[aa2][aax2]; 

				if(w >= cutoff) {
					if(!pair_families[i]) {
						pair_families[i] = (long*)malloc(MAX_PAIR_FAMILY * sizeof(long));
						pair_families[i][0] = 0;
					}
					idx = pair_families[i][0];
					pair_families[i][++idx] = j;
					pair_families[i][0] = idx;
					assert(idx < MAX_PAIR_FAMILY);
				}
			}
		}
	}
	return SUCCESS;
}

//------------------------------------------------------------
void hier_pair_primer_families_release(sparse_alignment_hier_data* _data)
{
	unsigned i, count = sizeof(_data->pair_families) / sizeof(_data->pair_families[0]);
	assert(_data);

	for(i = 0; i < count; i++) {
		if(_data->pair_families[i]) {
			free(_data->pair_families[i]);
			_data->pair_families[i] = NULL;
		}
	}
}

//------------------------------------------------------------------------------------
int hier_kernel_init_first_sequence(sparse_alignment_hier_data* _data)
{
	char*	 aas = _data->seq1->aas;
	unsigned l = _data->seq1->length - 1, idx, i, j, related_idx, count;
	//profile	 match_starts;
	long_pointer* pair_families =(long_pointer*)_data->pair_families; 
	long_pointer* pair_starts = (long_pointer*)_data->pair_starts;
	memset(pair_starts, 0, sizeof(_data->pair_starts));

	/*_data->match_starts = (profile)malloc((l + 1) * sizeof(coord_array20));
	memset(_data->match_starts, 0, sizeof(coord_array20));
	match_starts = _data->match_starts;

	// single praimers
	for(i = 0; i < l; i++) {
		match_starts[++match_starts[0][aas[i]]][aas[i]] = i; // indices start from 1
	}*/
	
	// pair primers
	for(i = 0; i < l; i++) {
		idx = aas[i] * 20 + aas[i + 1];
		if (idx<400){
			count = pair_families[idx] ? pair_families[idx][0]: 0;
			while(count) {
				related_idx = pair_families[idx][count--];
				if(!pair_starts[related_idx]) {
					pair_starts[related_idx] = (long*)malloc(MAX_PAIR_FAMILY * sizeof(long));
					assert(pair_starts[related_idx]);
					pair_starts[related_idx][0] = 0;
					}
				j = pair_starts[related_idx][0];
				pair_starts[related_idx][++j] = i;
				pair_starts[related_idx][0] = j;
				assert(j < MAX_PAIR_FAMILY);
			}
		}
	}
	return SUCCESS;
}
//------------------------------------------------------------
int hier_kernel_init_matchstarts(sparse_alignment_hier_data* _data)
{
	char*	 aas = _data->seq1->aas;
	unsigned l = _data->seq1->length - 1, i;
	profile	 match_starts;

	_data->match_starts = (profile)malloc((l + 1) * sizeof(coord_array20));
	memset(_data->match_starts, 0, sizeof(coord_array20));
	match_starts = _data->match_starts;

	// single praimers
	for(i = 0; i < l; i++) {
		match_starts[++match_starts[0][aas[i]]][aas[i]] = i; // indices start from 1
	}
	return SUCCESS;
}

//------------------------------------------------------------
int hier_kernel_init_first_sequence_Gen(sparse_alignment_hier_data* _data)
{
	char*	 aas = _data->seq1->aas;
	unsigned l = _data->seq1->length - 1, idx, i, j, related_idx;
	//profile	 match_starts;
	long_pointer* pair_starts = (long_pointer*)_data->pair_starts;
	memset(pair_starts, 0, sizeof(_data->pair_starts));

	/*_data->match_starts = (profile)malloc((l + 1) * sizeof(coord_array20));
	memset(_data->match_starts, 0, sizeof(coord_array20));
	match_starts = _data->match_starts;*/
	
	// pair primers
	for(i = 0; i < l; i++) {
		idx = aas[i] * 20 + aas[i + 1];
		if (idx<400){
			related_idx = idx;
			if(!pair_starts[idx]) {
				pair_starts[idx] = (long*)malloc(MAX_PAIR_FAMILY * sizeof(long));
				assert(pair_starts[idx]);
				pair_starts[idx][0] = 0;
			}
			j = pair_starts[idx][0];
			pair_starts[idx][++j] = i;
			pair_starts[idx][0] = j;
			assert(j < MAX_PAIR_FAMILY);
		}
	}
	return SUCCESS;
}

//------------------------------------------------------------
int hier_compute_pair_primer_kernels(
	sparse_alignment_hier_data* _data, coord_type _i, coord_type _j, coord_type _m, coord_type _n,
	kernel* _kernels, unsigned* _count//, kernel* kernels_dop_, unsigned* count_dop
)
{
	char *aas = _data->seq2->aas;
	unsigned i, l = _data->seq2->length, j, /*mask_size,*/ idx, pair_count;
	long_pointer* pair_starts = (long_pointer*)_data->pair_starts;
	//kernel krn = { 0 };

	// initialization
	char* mask_matrix = _data->mask_matrix;

	*_count = 0;
	_data->point_count = 0;
	j = _n;
	do {
		idx = aas[j] * 20 + aas[j + 1];
		if (idx<400) {
			pair_count = pair_starts[idx] ? pair_starts[idx][0] : 0;
			while(pair_count) {
				i = pair_starts[idx][pair_count--];
				if(i > _m) break;
				if(i >= _i && !mask_matrix[i * l + j]) {
					hier_check_hsp_arround_match(_data, i, j, _i, _j, _m, _n, _kernels, _count);
				}	
			}
		}
	} while(j-- > _j);

	return *_count;
}

//----------------------------------------------------------------------
// Searches HSPs over the Niddleman-Wunsch graph seq1 x seq2
// works with G_kernel_compute_data global structure
int hier_compute_single_primer_kernels(
	sparse_alignment_hier_data* _data, coord_type _i, coord_type _j, coord_type _m, coord_type _n,
	kernel* _kernels, unsigned* _count
)
{
	char symbol, *aas = _data->seq2->aas;
	unsigned i, l = _data->seq2->length, j, match_count/*, mask_size*/;
	profile	 match_starts = _data->match_starts;

	// initialization
	char* mask_matrix = _data->mask_matrix;
		
	*_count = 0;
	_data->point_count = 0;
	for(j = _j; j <= _n; j++) {
		symbol = *aas;
		match_count = (symbol<20) ? match_starts[0][symbol] : 0;
		while(match_count) {
			i = match_starts[match_count--][symbol];
			if(i > _m) break;
			if(i >= _i && !mask_matrix[i * l + j]) {
				hier_check_hsp_arround_match(_data, i, j, _i, _j, _m, _n, _kernels, _count);
			}
		}
		aas++;
	}

	return *_count;
}

//----------------------------------------------------------------------
// checks the (i,j)-match enter to the HSP 
score_type hier_check_hsp_arround_match(sparse_alignment_hier_data* _data, 
						coord_type i, coord_type j, 
						coord_type _i, coord_type _j, coord_type _m, coord_type _n, 
						kernel* _krn, unsigned* _count//, kernel* krn_dop_, unsigned* count_dop_
){
	char* seq1 = _data->seq1->aas;
	char* seq2 = _data->seq2->aas;
	int len1 = _m + 1;
	int len2 = _n + 1;
	unsigned	enuogh_kern = (unsigned)sqrt((_m-_i+_n -_j)/2) /2 ;
	coord_type i_forward = i - 1, i_backward = i;
	coord_type j_forward = j - 1, j_backward = j;
	coord_type i_max_forward = i, j_max_forward = j;
	coord_type i_max_backward = i, j_max_backward = j, l = 0;
	score_type height, w_max = 0, w_max0 = 0, reminder, hole = _data->current_kernel_params->max_hole;
	char better = 0, no_back = 0, no_fore = 0;
	score_type min_score = _data->current_kernel_params->min_score;

	height = 0;
	do {
		l = __min(len1 - i_forward - 1, len2 - j_forward - 1);
		while(height >= 0 && l--) {
			height += subst_matrix[seq1[++i_forward]][seq2[++j_forward]];
			if(height > w_max) { w_max = height; i_max_forward = i_forward; j_max_forward = j_forward; better = 1; }
			else if(w_max - height >= hole) {
				no_fore = 1;
				break;
			}
		}
		if(!better || no_back) break;
		reminder = height;
		height = w_max - w_max0;
		w_max0 = w_max - reminder;
		better = 0;
		l = __min(i_backward - _i, j_backward - _j);
		while(height >= 0 && l--) {
			height += subst_matrix[seq1[--i_backward]][seq2[--j_backward]];
			if(height > w_max) { w_max = height; i_max_backward = i_backward; j_max_backward = j_backward; better = 1; }
			else if(w_max - height >= hole) {
				no_back = 1;
				break;
			}
		}
		if(!better || no_fore) break;
		reminder = height;
		height = w_max - w_max0;
		w_max0 = w_max - reminder;
		better = 0;
	} while(1);

	// keep track in mask_matrix
	i = i_backward; j = j_backward;
	while(i <= i_forward) {
		assert(i * len2 + j >= 0 && i * len2 + j < len1 * len2);
		_data->mask_matrix[i++ * _data->seq2->length + j++] = 1;
	}

	// save HSP
	if(w_max >= min_score) {
		assert(*_count < MAX_KERNELS);
		_krn[*_count].id = i;
		_krn[*_count].x = i_max_backward;
		_krn[*_count].y = j_max_backward;
		_krn[*_count].length = i_max_forward - i_max_backward + 1;
		_krn[*_count].score = w_max;
		_data->point_count += _krn[*_count].length;
		(*_count) ++ ;
	} /*else 
		if ((*_count) < enuogh_kern && w_max > min_score/2){
			assert(Dop_kernel_count < MAX_KERNELS);
			Kernels_dop[Dop_kernel_count].id = i;
			Kernels_dop[Dop_kernel_count].x = i_max_backward;
			Kernels_dop[Dop_kernel_count].y = j_max_backward;
			Kernels_dop[Dop_kernel_count].length = i_max_forward - i_max_backward + 1;
			Kernels_dop[Dop_kernel_count].score = w_max;
			//_data->point_count += _krn[*count_dop_].length;
			(Dop_kernel_count) ++ ;
		}*/
	return w_max;
}

//------------------------------------------------------------------------------------
void hier_kernel_init_first_sequence_release(sparse_alignment_hier_data* _data)
{
	unsigned i, count = sizeof(_data->pair_starts) / sizeof(_data->pair_starts[0]);

	// pair primers
	for(i = 0; i < count; i++) {
		if(_data->pair_starts[i]) {
			free(_data->pair_starts[i]);
			_data->pair_starts[i] = NULL;
		}
	}

	// signle primers
	//if (_data->match_starts) 
		free(_data->match_starts);
	_data->match_starts = NULL;
}

//------------------------------------------------------------------------------------
int hier_hexograph_build(sparse_alignment_hier_data* _data)
{
	int res, i, m = _data->seq1->length, n = _data->seq2->length;

	_data->current_iter = -1;
	for(i = 0; i < m; i++) _data->path[i] = -1;
	_data->mask_matrix = (char*)malloc(m * n * sizeof(char));
	Dop_kernel_count = 0;

	res = iter_hexograph_build(_data, 0, 0, m - 1, n - 1, 1,1);

	free(_data->mask_matrix);
	_data->mask_matrix = 0;


	return res;
}

//------------------------------------------------------------------------------------------------------------
int iter_hexograph_build(sparse_alignment_hier_data* _data, coord_type _i, coord_type _j, coord_type _m, coord_type _n,
						 char _free_begins, char _free_ends)
{
	int i, j, m = _data->seq1->length, n = _data->seq2->length, *path = _data->path /*, prev_diag = 0*/, prev_kernel = 0;

	//     
	//     Sqrt(max(_m - _i, _n - _j))
	kernel*			kernels; 
	unsigned		count, add_count, calc_count = (unsigned)sqrt((_m-_i+_n -_j+2)/2), mask_size = m * n * sizeof(char); 
	kernel*			current_kernel; 
	kernel_list				open_kernels = { 0 };
	kernel_list_node*		open_krn_iterator = NULL;
	hexograph_node*  prev_node = 0, *current_node, *max_node = NULL; //  
	hexograph_node   *prev_top_nodes[2048];//[1024];//;  
	hexograph_node   *diag_nodes[4096];//[2048]; 
	hexograph_node** prev_diag_nodes = (hexograph_node**)diag_nodes;
	score_type		 max_score = 0;
	hexograph		 current_hexograph;
	char			 do_impruve = _data->do_impruve;


	_data->current_iter ++;
	/*if(_m - _i + 1 < SMALLEST_SIZE_ITER || _n - _j + 1 < SMALLEST_SIZE_ITER )
		// || _data->current_iter >= _data->iter_count) 
	{
		if (do_impruve) hier_hexograph_sw_doub_improve(_data, _i , _j, _m, _n, _free_begins, _free_ends);
		return SUCCESS;
	}*/

	//  current_kernel_params
	_data->current_kernel_params = &(_data->krn_params[_data->current_iter]); 

	memset(_data->mask_matrix, 0, mask_size);
	kernels = (kernel*)malloc(MAX_KERNELS * sizeof(kernel));
	if (!kernels) return 0;// E_OUTOFMEMORY;

	/*if (_data->krn_params[0].ker_method == 3) 
		total_count = hier_compute_single_primer_kernels(_data, _i, _j, _m, _n, kernels, &count);
	else total_count = */
	
	//calc_count = (calc_count > 5)? calc_count: 5;	
	//calc_count = 100;
	Dop_kernel_count = 0;
	hier_compute_pair_primer_kernels(_data, _i, _j, _m, _n, kernels, &count);
	
	//debug_print_kernels("kernels.txt", kernels, count);

	if(calc_count < count) {
		count = reget_kernels(kernels, count, calc_count);
		_data->point_count = reget_point_count(kernels, count);
	} /**//*else {
		if (count < calc_count/2 + 1) {
			add_count = reget_kernels(Kernels_dop, Dop_kernel_count, calc_count/2 +1 - count);
			for (i=0; i<add_count; i++)	{
				assert(count < MAX_KERNELS);
				kernels[count].id = Kernels_dop[i].id;
				kernels[count].x  = Kernels_dop[i].x ;
				kernels[count].y  = Kernels_dop[i].y ;
				kernels[count].length = Kernels_dop[i].length;
				kernels[count].score  = Kernels_dop[i].score ;
				_data->point_count += kernels[count].length;
				count++;
			}
		}
	}*/
	//debug_print_kernels("kernels.txt", kernels, count);

	//       !
	//assert(count>0);
	if (count==0) {

		if (kernels) {	free(kernels);	kernels = NULL;}
		if (do_impruve) hier_hexograph_sw_doub_improve(_data, _i , _j, _m, _n, _free_begins, _free_ends); //hier_hexograph_sw_doub_improve(_data, _i-1 , _j-1, _m, _n, _free_begins, _free_ends);
		//if (do_impruve) iter_hexograph_build(_data,_i, _j, _m , _n,	_free_begins, _free_ends);
		return SUCCESS; //0;
	}

	kernels = (kernel*)realloc(kernels, count*sizeof(kernel));
	//debug_print_kernels("kernels.txt", kernels, count);

	//    x   y
	qsort(kernels, count, sizeof(kernel), kernel_compare);
	//debug_print_kernels("kernels.txt", kernels, count);

	_data->owners_tail = _data->owners;

	//  
	current_hexograph.pool_size = _data->point_count + 1;
	current_hexograph.node_pool = (hexograph_node*)malloc(current_hexograph.pool_size * sizeof(hexograph_node));
	
	assert(current_hexograph.node_pool);
	memset(current_hexograph.node_pool, 0, current_hexograph.pool_size * sizeof(hexograph_node));
	current_node = current_hexograph.node_pool;
	
	memset(prev_top_nodes, 0, m * sizeof(hexograph_node*));
	memset(prev_diag_nodes, 0, (m + n) * sizeof(hexograph_node*));
	//        ,  
	prev_diag_nodes += n;

	current_kernel = kernels;
	kernels += count;
	//   
	for(j = _j; j <= _n || open_kernels.head != NULL; j++) {

		// j -  
		open_krn_iterator = open_kernels.head;
		//       
		while(current_kernel != kernels && current_kernel->y == j) {
			while(open_krn_iterator && current_kernel->x > open_krn_iterator->data.x)
				open_krn_iterator = open_krn_iterator->next;
			//   open_krn_iterator
			kernel_list_insert(&open_kernels, current_kernel, open_krn_iterator);
			current_kernel ++ ;
		}

		prev_node = 0;
		open_krn_iterator = open_kernels.head;
		while(open_krn_iterator) {
			
			i = open_krn_iterator->data.x;
			current_node->point_of_kernel = open_krn_iterator->data.id;

			current_node->left = prev_node;
			prev_node = current_node;

			current_node->prev = prev_diag_nodes[i - j];
			prev_diag_nodes[i - j] = current_node;

			current_node->top = prev_top_nodes[i];
			prev_top_nodes[i] = current_node;

			current_node->x_coord = i; current_node->y_coord = j; current_node->diag = i - j;

			if(open_krn_iterator->start) hier_compute_hexograph_node_inner(_data,current_node);//(&current_hexograph, current_node); 
			else hier_compute_hexograph_node_start(_data,current_node);//(&current_hexograph, current_node); 

			//debug_dump_hexograph(G_debug_dump, current_node);

			if(--open_krn_iterator->data.length) {
				open_krn_iterator->data.x++;
				open_krn_iterator->data.y++;
				open_krn_iterator->start = 1; // 
				open_krn_iterator = open_krn_iterator->next;
			} else {
				//  
				open_krn_iterator = kernel_list_delete(&open_kernels, open_krn_iterator);

				if(current_node->strong) {
					
					*(_data->owners_tail) = current_node;
					_data->owners_tail ++ ;
					if(current_node->best_score >= max_score) {
						max_score = current_node->best_score;
						max_node = current_node;
					}
				}
			}

			current_node ++ ;

			if(current_node - current_hexograph.node_pool >= current_hexograph.pool_size ) {
				current_hexograph.pool_size += POOL_SIZE;
				current_hexograph.node_pool = (hexograph_node*)realloc(current_hexograph.node_pool, current_hexograph.pool_size * sizeof(hexograph_node));
			}
			if(!open_krn_iterator) break;
		}
	}

	current_hexograph.pool_tail = current_node;

	if(!max_node) {
		if (do_impruve) hier_hexograph_sw_doub_improve(_data, _i , _j, _m, _n, _free_begins, _free_ends);
		hexograph_release(&current_hexograph);
		kernels -= count;	
		if (kernels) {	free(kernels);	kernels = NULL;}
		_data->current_iter --;
		return 0;
	}

	//  
	
	if (do_impruve)
		 if ((_data->current_iter +1 < _data->iter_count) &&
			 ((_m - max_node->x_coord) > SMALLEST_SIZE_ITER) && ((_n - max_node->y_coord) > SMALLEST_SIZE_ITER) ) 
			   iter_hexograph_build(_data, max_node->x_coord + 1, max_node->y_coord + 1, _m, _n, 0, _free_ends);
		 else 
			 hier_hexograph_sw_doub_improve(_data,  max_node->x_coord+1 , max_node->y_coord+1, _m, _n, 0, _free_ends);
		      //hexograph_path_improve(max_node->x_coord,max_node->y_coord, _m+1, _n+1,_data->seq1->aas, _data->seq2->aas, path); 
			  //hexograph_path_improve_lokal(max_node->x_coord,max_node->y_coord, _m+1, _n+1,_data->seq1->aas, _data->seq2->aas, path, 0, _free_ends);
	/**/
	while(1) {
		//    
		i = max_node->x_coord;
		j = max_node->y_coord;
		path[i] = j;
		prev_kernel = max_node->point_of_kernel;
		// prev_diag = max_node->diag;
		max_node = max_node->best_link;
		if(!max_node) break;
		if(/*prev_diag == max_node->diag &&*/ prev_kernel == max_node->point_of_kernel) {
			while(i > max_node->x_coord && j > max_node->y_coord) path[i--] = j--;
		} else {
			if(i == max_node->x_coord + 1 || j == max_node->y_coord + 1)
				while(--i > max_node->x_coord) path[i] = -1;
			else {
				// hexograph_sw_improve(max_node->x_coord, max_node->y_coord, i, j, 0, 0);
				if (do_impruve) 
					if ((_data->current_iter +1 < _data->iter_count) &&
						(i - max_node->x_coord > SMALLEST_SIZE_ITER) && (j - max_node->y_coord > SMALLEST_SIZE_ITER )) 
							iter_hexograph_build(_data, max_node->x_coord + 1, max_node->y_coord + 1, i, j, 0, 0);
					else 
						hier_hexograph_sw_doub_improve(_data,  max_node->x_coord+1 , max_node->y_coord+1, i-1, j-1, 0, 0);
						 //hexograph_path_improve(max_node->x_coord,max_node->y_coord, i+1, j+1,_data->seq1->aas, _data->seq2->aas, path); 
						//hexograph_path_improve_lokal(max_node->x_coord,max_node->y_coord, i+1, j+1,_data->seq1->aas, _data->seq2->aas, path, 0, 0);
			}
		}
	}

	// hexograph_sw_improve(-1, -1, i, j, 1, 0);
	if (do_impruve)
		if ((_data->current_iter +1 < _data->iter_count) &&
			(i-_i+1 > SMALLEST_SIZE_ITER) && (j-_j+1 > SMALLEST_SIZE_ITER )) 
				iter_hexograph_build(_data, _i, _j, i, j, _free_begins , 0);
		else 
			hier_hexograph_sw_doub_improve(_data, _i , _j, i-1, j-1, _free_begins, 0);
			 //hexograph_path_improve(_i-1,_j-1, i+1, j+1,_data->seq1->aas, _data->seq2->aas, path); 
			 //hexograph_path_improve_lokal(_i-1,_j-1, i+1, j+1,_data->seq1->aas, _data->seq2->aas, path, _free_begins, 0);
		
	_data->max_score = max_score;
	hexograph_release(&current_hexograph);
	
	kernels -= count;	
	if (kernels) {	free(kernels);	kernels = NULL;}

	_data->current_iter --;
	return SUCCESS;
}


//--------------------------------------------------
char hier_compute_hexograph_node_start(sparse_alignment_hier_data* _data, hexograph_node* _node)
{
	hexograph_node** iterator = _data->owners;
	hexograph_node*	 max_owner = NULL, *left, *top, *jump, *owner;
	coord_type diag = _node->diag, diff;
	score_type max_owner_weight = 0;
	score_type w_left = 0, w_top = 0, w_diag, w, w_jump;
	score_type kcp = _data->KCP[_data->current_iter], gep = _data->KEP[_data->current_iter];
	char strong = 1;

	assert(_node);

	//  
//	max_owner = zvi_galil_lookup_owner(_node);
//	max_owner_weight = max_owner->jump_score - diff * gep;

	while(iterator != _data->owners_tail) {
		diff = ABS((*iterator)->diag - diag);
		if((*iterator)->x_coord < _node->x_coord && (*iterator)->y_coord < _node->y_coord) {
			if((w = (*iterator)->jump_score - diff * gep) >= max_owner_weight) {
				max_owner_weight = w;
				max_owner = *iterator;
			}
		}
		iterator ++;
	}
	// 

	//      
	owner = left = _node->left;
	while(owner) {
		if(owner->strong && owner->y_coord < _node->y_coord) {
			diff = ABS(diag - owner->diag);
			if((w = owner->jump_score - diff * gep) > max_owner_weight) {
				max_owner_weight = w;
				max_owner = owner;
				break;
			}
		}
		owner = owner->best_link;
	}
	owner = top = _node->top;
	while(owner) {
		if(owner->strong && owner->x_coord < _node->x_coord) {
			diff = ABS(diag - owner->diag);
			if((w = owner->jump_score - diff * gep) > max_owner_weight) {
				max_owner_weight = w;
				max_owner = owner;
				break;
			}
		}
		owner = owner->best_link;
	}

	if(max_owner) {
		if(max_owner_weight < 0) {
			max_owner_weight = 0;
			max_owner = 0;
		}
	}
	
	w_diag = max_owner_weight + subst_matrix[_data->seq1->aas[_node->x_coord]][_data->seq2->aas[_node->y_coord]];

	if(left) {
		w_left = left->jump_score - gep * ABS(left->diag - diag);
		left = left->jump_link;
	};
	if(top) {
		w_top = top->jump_score - gep * ABS(top->diag - diag);
		top = top->jump_link;
	};

	if(w_top > w_left) { w_jump = w_top; jump = top; }
	else { w_jump = w_left; jump = left; }
	
	if(w_diag > w_jump) {
		_node->best_score = w_diag;
		_node->best_link = max_owner;

		if(w_diag - kcp >= w_jump) { 
			_node->jump_score = w_diag - kcp;
			_node->jump_link = max_owner;
		} else { 
			_node->jump_score = w_jump;
			_node->jump_link = jump ? jump->jump_link : NULL;
		};
	} else {
		_node->best_score = _node->jump_score = w_jump;
		_node->best_link = _node->jump_link = jump;
		strong = 0;
	}
	return (_node->strong = strong);
}

//--------------------------------------------------
char hier_compute_hexograph_node_inner(sparse_alignment_hier_data* _data, hexograph_node* _node)
{
	score_type kcp = _data->KCP[_data->current_iter], gep = _data->KEP[_data->current_iter];
	score_type w_left = 0, w_top = 0, w_jump, w;
	hexograph_node* left = _node->left, *top = _node->top, *jump = NULL, *prev = _node->prev;
	score_type w_diag = prev->best_score + subst_matrix[_data->seq1->aas[_node->x_coord]][_data->seq2->aas[_node->y_coord]];
	coord_type diag = _node->diag;
	char strong = 1;

	assert(_node);

	if(left) {
		w_left = left->jump_score - gep * ABS(left->diag - diag);
		left = left->strong ? left : left->jump_link;
	};
	if(top) {
		w_top = top->jump_score - gep * ABS(top->diag - diag);
		top = top->strong ? top : top->jump_link;
	};
	if(w_left > w_top) { w_jump = w_left; jump = left; }
	else { w_jump = w_top; jump = top; };
	if(prev->jump_score > w_jump) { w_jump = prev->jump_score; jump = prev->jump_link;};
	
	if(w_diag >= w_jump) {
		_node->best_score = w_diag;
		_node->best_link = prev->strong ? prev : prev->best_link;

		if((w = w_diag - kcp) >= w_jump) {
			_node->jump_score = w;
			_node->jump_link = _node /*prev->strong ? prev : prev->best_link*/; //   
		} else {
			_node->jump_score = w_jump;
			_node->jump_link = jump ? jump->jump_link : NULL;
		}

		if(_node->jump_score < prev->jump_score) {
			_node->jump_score = prev->jump_score;
			_node->jump_link = prev->jump_link;
			strong = 0;
		}
	} else {

		_node->best_score = _node->jump_score = w_jump;
		_node->best_link = _node->jump_link = jump ? jump->jump_link : NULL;
		
		if(_node->jump_score < prev->jump_score) {
			_node->jump_score = prev->jump_score;
			_node->jump_link = prev->jump_link;
		}
		strong = 0;
	}
	return (_node->strong = strong);
}

//-----------------------------------------------------------------------------
int hier_hexograph_sw_doub_improve (sparse_alignment_hier_data* _data,
						coord_type _i1, coord_type _j1,
						coord_type _i2, coord_type _j2,
						char _free_begins, char _free_ends 
						//long* num_nothing_, long* num_usefull_
)
{
	int				m = _i2 - (_i1)+1, n = _j2 - (_j1)+1;//m = _i2 - (++_i1), n = _j2 - (++_j1);
	int				i, j, i_max = -1, j_max = -1;
	score_type		GEP = _data->sw_imp_GEP, GOP = _data->sw_imp_GOP,  w_max, jump_left = 0, jump_top = 0;
	sw_link			*sw_links;
	sw_cell			*current_line, *prev_line, *p = NULL;
//	sw_cell			*current_line = G_sw_compute_data.line1, *prev_line = G_sw_compute_data.line1, *p = NULL;

	char			dir, open_flag;

	char*			seq2 = _data->seq2->aas, *seq1 = _data->seq1->aas; 
	int*			path = _data->path;

	int				direct_nat;
  extern FILE*       f_imp_deb;
  int empty = (int) ((1-_free_begins) * (1-_free_ends));
  int edge =  (int) (_free_begins + _free_ends);

  assert(seq1);
	assert(seq2);

	if(m <= 0 || n <= 0) {
		w_max = -99999;
		goto FIN_PRINT;
		return SUCCESS;
	}

	// general memory allocation for NW-graph
	current_line = (sw_cell*)malloc((m + 1) * sizeof(sw_cell));
	assert(current_line);
	prev_line = (sw_cell*)malloc((m + 1) * sizeof(sw_cell));
	assert(prev_line);

	sw_links = (sw_link*)malloc(m * n * sizeof(sw_link));
	assert(sw_links);
	memset(sw_links, 0, m * n * sizeof(sw_link));
	
	if(_free_begins) {
		memset(prev_line, 0, (m + 1) * sizeof(sw_cell));
		memset(current_line, 0, (m + 1) * sizeof(sw_cell));//memset(current_line, 0, sizeof(sw_cell));
	} else {
		prev_line[0].diag = 0;
		for(i = 1; i <= m; i++) {
			prev_line[i].jump = -10000;
			prev_line[i].diag = - GOP - GEP * i;
		}
	}
	w_max = (score_type)- 10000 * (1-_free_ends);

	current_line++;
	prev_line++;

	//   
	for(j = 0; j < n; j++)
	{
		//if(!_free_begins) {
			current_line[-1].jump = (score_type) -10000 * (1 - _free_begins);
			current_line[-1].diag = (- GOP - GEP * (j + 1)) * (1 - _free_begins);
		//}

		//  
		for(i = 0; i < m; i++) {

			current_line[i].diag = prev_line[i-1].diag + subst_matrix[seq1[_i1 + i]][seq2[_j1 + j]];
														
			// 5-comparison goes here
			dir = 0; open_flag = 0;

			if(current_line[i-1].jump < current_line[i-1].diag - GOP) {
				open_flag |= DIR_LEFTOPEN;
				jump_left = current_line[i-1].diag - GOP - GEP;
			} else 
				jump_left = current_line[i-1].jump - GEP;

			if(prev_line[i].jump < prev_line[i].diag - GOP) {
				open_flag |= DIR_TOPOPEN;
				jump_top = prev_line[i].diag - GOP - GEP;
			} else 
				jump_top = prev_line[i].jump - GEP;

			if(jump_top > jump_left) {
				current_line[i].jump = jump_top;
				dir = DIR_TOP;
			} else {
				current_line[i].jump = jump_left;
				dir = DIR_LEFT;
			}

			if(current_line[i].jump > current_line[i].diag) {
				current_line[i].diag = current_line[i].jump;
			} else {
				dir = DIR_DIAG;
			}

			if(_free_begins && (current_line[i].diag < 0)) {
				//if(current_line[i].diag < 0) {
					current_line[i].diag = 0;
					dir = DIR_TERMINATOR;
				//}
			}
			//if(_free_ends) {
				if(current_line[i].diag >= w_max) {
					w_max = current_line[i].diag;
					i_max = i; j_max = j;
				}
			//}

			sw_links[j * m + i].link = dir;
			sw_links[j * m + i].open_flag = open_flag;
		}
		// swap prev/current
		p = prev_line;
		prev_line = current_line;
		current_line = p;
	}

	if (!_free_ends) w_max = prev_line[m-1].diag;
	
	if(w_max < -25){ // -25 impruvment !!! (- GOP - GEP*abs(m-n+1))
		free(sw_links);
		free(--current_line);
		free(--prev_line);
		(Sw_nothing) += 1;

		goto FIN_PRINT;

		return SUCCESS;  
	}/**/

	//(Sw_usefull) += 1;
	//  ,    
	if(_free_ends) {
		for(i = m - 1; i > i_max; path[_i1 + i--] = -1);
		i = i_max; j = j_max;
	} else {
		i = m - 1; 
		j = n - 1;
	}
	while(i >= 0 && j >= 0) {
		switch(sw_links[j * m + i].link) {
		case DIR_DIAG:
			path[_i1 + i--] = _j1 + j--;
			break;
		case DIR_TOP :
			do {
				if(sw_links[j * m + i].link == DIR_LEFT || j<0) break;
				 //assert(j >= 0);
				path[_i1 + i] = _j1 + j; 
			} while (!(sw_links[(j--) * m + i].open_flag & DIR_TOPOPEN));
			break;
		case DIR_LEFT :
			do {
				if(sw_links[j * m + i].link == DIR_TOP || i<0) break;
				 //assert(i >= 0);
				path[_i1 + i] = -1;
				direct_nat = DIR_LEFTOPEN;
			} while (!(sw_links[j * m + i--].open_flag & DIR_LEFTOPEN));
			break;
		default:
			assert((sw_links[j * m + i].link == DIR_TERMINATOR) && _free_begins);
			j = -1; // in order to exit while !!!
			break;
		}
	}

	if(_free_begins) {
		while(i >= 0) path[_i1 + i--] = -1;
	}

	free(sw_links);
	free(--current_line);
	free(--prev_line);

FIN_PRINT:
        /*fprintf(f_imp_deb, "%s\t%s\t",
         _data->seq1->name, _data->seq2->name  // 11.
        );

        fprintf(f_imp_deb, "%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t",
           m, n,
		   _i1,_i2,_j1,_j2, 
		   _free_begins,  _free_ends,
           empty, edge

        );
        fprintf(f_imp_deb, "%f\t\n",
           w_max
        );*/


	return SUCCESS;
}