

#ifndef _hier_h_
#define _hier_h_

///////////////////////////////////////////////////////////////
//
//	File:		hier.h
//	Created:	20.12.2001
//	Purpose:	Iterative Sparse dynamic programing in sequence alignment
//
//	Comments:	
//	Authors:	Bogopolsky Gennady, EIMB (c) 2002
//
///////////////////////////////////////////////////////////////

#define		MAX_HIER				4
#define		SMALLEST_SIZE_ITER		8

#include "sparse.h"

typedef struct _sparse_alignment_hier_data
{
	sequence*	seq1;
	sequence*	seq2;

	kernel_params	krn_params[MAX_HIER];
	//kernel_params	max_kernel_params;
	//kernel_params	min_kernel_params;
	kernel_params*	current_kernel_params;
	unsigned		point_count;

	score_type	KCP[MAX_HIER];
	score_type	KEP[MAX_HIER];
	score_type	sw_imp_GOP;
	score_type	sw_imp_GEP;
	int			iter_count, current_iter;

	hexograph_node*		owners[POOL_SIZE];
	hexograph_node**	owners_tail;

	int			path[MAX_LEN];
	score_type	max_score;

	char*		mask_matrix;
	// data for per-query initialization
	long_pointer	pair_starts[400];
	long_pointer	pair_families[400];	

	profile			match_starts;
	char			do_impruve;

} sparse_alignment_hier_data;

int hier_init_pair_primer_families (sparse_alignment_hier_data* _data);
int hier_kernel_init_first_sequence(sparse_alignment_hier_data* _data);
int hier_kernel_init_first_sequence_Gen(sparse_alignment_hier_data* _data);
int hier_kernel_init_matchstarts(sparse_alignment_hier_data* _data);

int hier_compute_pair_primer_kernels(sparse_alignment_hier_data* _data, coord_type _i, coord_type _j, coord_type _m, coord_type _n, kernel* _kernels, unsigned* _count);
int hier_compute_single_primer_kernels(sparse_alignment_hier_data* _data, coord_type _i, coord_type _j, coord_type _m, coord_type _n, kernel* _kernels, unsigned* _count);
score_type hier_check_hsp_arround_match(sparse_alignment_hier_data* _data, coord_type i, coord_type j, coord_type _i, coord_type _j, coord_type _m, coord_type _n, kernel* _krn, unsigned* _count);

void hier_kernel_init_first_sequence_release(sparse_alignment_hier_data* _data);
void hier_pair_primer_families_release(sparse_alignment_hier_data* _data);

int  iter_hexograph_build(sparse_alignment_hier_data* _data, coord_type _i, coord_type _j, coord_type _m, coord_type _n,char _free_begins, char _free_ends);
int  hier_hexograph_build(sparse_alignment_hier_data* _data);
char hier_compute_hexograph_node_start(sparse_alignment_hier_data* _data, hexograph_node* _node);
char hier_compute_hexograph_node_inner(sparse_alignment_hier_data* _data, hexograph_node* _node);

int hier_hexograph_sw_doub_improve (sparse_alignment_hier_data* _data,
						coord_type _i1, coord_type _j1,
						coord_type _i2, coord_type _j2,
						char _free_begins, char _free_ends);

void hexograph_path_improve(coord_type _i1, coord_type _j1,
							coord_type _i2, coord_type _j2,
							char* _seq1, char* _seq2, int* _path);
void hexograph_path_improve_lokal(coord_type _i1, coord_type _j1,
							coord_type _i2, coord_type _j2,
							char* _seq1, char* _seq2, int* _path
							,char free_begin_, char free_end_);

//void mysort(kernel* _kernels, unsigned _count);
//int  kernel_compare(const void * arg1, const void * arg2);
//unsigned reget_kernels(kernel* _kernels, unsigned _count, unsigned _required);
kernel_list_node* kernel_list_insert(kernel_list* _list, kernel* _krn, kernel_list_node* _before);
kernel_list_node* kernel_list_delete(kernel_list* _list, kernel_list_node* _node);

long	Sw_nothing, Sw_usefull;


#endif _hier_h_