
///////////////////////////////////////////////////////////////
//
//	File:		kernels.c
//	Created:	20.12.2001
//	Purpose:	All about kernels and their's search methods
//
//	Comments:	
//	Authors:	Bogopolsky Gennady, EIMB (c) 2001
//
///////////////////////////////////////////////////////////////


#include "skoplib.h"
#include "windows.h"


// Global definitions
kernel_compute_data		G_kernel_compute_data;


score_type check_hsp_arround_match(coord_type i, coord_type j);
void debug_dump_kernels(FILE* _file, kernel_compute_data* _data);

//----------------------------------------------------------------------
int init_compute_kernels(
	sequence _seq1
)
{
	unsigned i, l = _seq1.length, idx;
	char*    aas = _seq1.aas;
	profile	 match_starts;

	G_kernel_compute_data.params->method = kernel_method_matches;

	G_kernel_compute_data.match_starts = (profile)malloc((l + 1) * sizeof(coord_array20));
	memset(G_kernel_compute_data.match_starts, 0, sizeof(coord_array20));
	match_starts = G_kernel_compute_data.match_starts;

	for(i = 0; i < l; i++) {
		idx = match_starts[0][aas[i]];
		match_starts[idx+1][aas[i]] = i; // indices start from 1
		match_starts[0][aas[i]] ++ ;
	}

	return SUCCESS;
}

//----------------------------------------------------------------------
// Searches HSPs over the Niddleman-Wunsch graph seq1 x seq2
// works with G_kernel_compute_data global structure
int compute_kernels(
	sequence* _seq1,
	sequence* _seq2, 
	kernel_params* _params
)
{
	char symbol, *aas = _seq2->aas;
	unsigned i, l = _seq2->length, j, match_count, mask_size;
	profile	 match_starts = G_kernel_compute_data.match_starts;

	// initialization
	G_kernel_compute_data.mask_matrix = (char*)malloc(mask_size = _seq1->length * _seq2->length * sizeof(char));
	memset(G_kernel_compute_data.mask_matrix, 0, mask_size);
	G_kernel_compute_data.seq1 = _seq1;
	G_kernel_compute_data.seq2 = _seq2;
	G_kernel_compute_data.params = _params;
	G_kernel_compute_data.kernel_count = 0;
	G_kernel_compute_data.coverage = 0;
	G_kernel_compute_data.strong_coverage = 0;
	G_kernel_compute_data.total_length = 0;
	G_kernel_compute_data.used_length = 0;
		
	for(i = 0; i < l; i++) {
		symbol = *aas;
		match_count = match_starts[0][symbol];
		while(match_count) {
			j = match_starts[match_count--][symbol];
			if(!G_kernel_compute_data.mask_matrix[j * l + i])
				check_hsp_arround_match(j, i);
		}
		aas ++ ;
	}

	G_kernel_compute_data.seq1 = G_kernel_compute_data.seq2 = NULL;

	return G_kernel_compute_data.kernel_count;
}

//----------------------------------------------------------------------
// checks the (i,j)-match enter to the HSP 
score_type check_hsp_arround_match(coord_type i, coord_type j)
{
	char* seq1 = G_kernel_compute_data.seq1->aas;
	char* seq2 = G_kernel_compute_data.seq2->aas;
	int len1 = G_kernel_compute_data.seq1->length;
	int len2 = G_kernel_compute_data.seq2->length;
	coord_type i_forward = i - 1, i_backward = i;
	coord_type j_forward = j - 1, j_backward = j;
	coord_type i_max_forward = i, j_max_forward = j;
	coord_type i_max_backward = i, j_max_backward = j, l = 0;
	score_type height, w_max = 0, w_max0 = 0, reminder, hole = G_kernel_compute_data.params->max_hole;
	char better = 1, no_back = 0, no_fore = 0;

	height = 0;
	do {
		l = __min(len1 - i_forward - 1, len2 - j_forward - 1);
		while(height >= 0 && l--) {
			height += subst_matrix[seq1[++i_forward]][seq2[++j_forward]];
			if(height > w_max) { w_max = height; i_max_forward = i_forward; j_max_forward = j_forward; better = 1; }
			else if(w_max - height >= hole) {
				no_fore = 1;
				break;
			}
		}
		if(!better || no_back) break;
		reminder = height;
		height = w_max - w_max0;
		w_max0 = w_max - reminder;
		better = 0;
		l = __min(i_backward, j_backward);
		while(height >= 0 && l--) {
			height += subst_matrix[seq1[--i_backward]][seq2[--j_backward]];
			if(height > w_max) { w_max = height; i_max_backward = i_backward; j_max_backward = j_backward; better = 1; }
			else if(w_max - height >= hole) {
				no_back = 1;
				break;
			}
		}
		if(!better || no_fore) break;
		reminder = height;
		height = w_max - w_max0;
		w_max0 = w_max - reminder;
		better = 0;
	} while(1);

	// keep track in mask_matrix
	/*i = i_backward; j = j_backward;
	while(i <= i_forward) {
		assert(i * len2 + j >= 0 && i * len2 + j < len1 * len2);
		G_kernel_compute_data.mask_matrix[i++ * len2 + j++] = 1;
	}*/
	// keep track in mask_matrix only kerles
	i = i_max_backward; j = j_max_backward;
	while(i <= i_max_forward) {
		assert(i * len2 + j >= 0 && i * len2 + j < len1 * len2);
		G_kernel_compute_data.mask_matrix[i * len2 + j] = 1;
		i++; j++;
	}/**/

	// save HSP
	if(w_max >= G_kernel_compute_data.params->min_score) {
		i = G_kernel_compute_data.kernel_count;
		assert(i < MAX_KERNELS*20);
		G_kernel_compute_data.kernels[i].id = i;
		G_kernel_compute_data.kernels[i].x = i_max_backward;
		G_kernel_compute_data.kernels[i].y = j_max_backward;
		G_kernel_compute_data.kernels[i].length = i_max_forward - i_max_backward + 1;
		G_kernel_compute_data.kernels[i].score = w_max;
		G_kernel_compute_data.kernel_count ++ ;
		G_kernel_compute_data.used_length += i_max_forward - i_max_backward + 1;
	}

	// DEBUG purpose
	G_kernel_compute_data.coverage += (i_forward - i_backward + 1);
	G_kernel_compute_data.total_length += (i_max_forward - i_max_backward + 1);
	// if(strong_segment) G_kernel_compute_data.strong_coverage += (i_forward - i_backward + 1);

	return w_max;
}

//------------------------------------------------------------
// release all kernel stuff
void kernel_compute_data_release(kernel_compute_data* _data)
{
	free(_data->mask_matrix);
	_data->mask_matrix = NULL;
}

//------------------------------------------------------------
// release all kernel stuff
void kernel_init_data_release(kernel_compute_data* _data)
{
	// release all stuff
	free(_data->match_starts);
	_data->match_starts = NULL;
}

//------------------------------------------------------------
int init_pair_primer_families(kernel_compute_data* _data)
{
	register int	i, j;
	unsigned		idx;
	score_type		w, w1, w2;
	char			aa1, aa2, aax1, aax2;
	score_type		cutoff = G_kernel_compute_data.params->pair_cutoff;
	long_pointer* pair_families = (long_pointer*)G_kernel_compute_data.pair_primer_data.pair_families;

	assert(_data);
	memset(G_kernel_compute_data.pair_primer_data.pair_families, 0, sizeof(G_kernel_compute_data.pair_primer_data.pair_families));

	for(i = 0; i < 400; i++) {
		for(j = 0; j < 400; j++) {
			aa1 = i / 20;
			aa2 = i % 20;
			aax1 = j / 20;
			aax2 = j % 20;
			w1 = subst_matrix[aa1][aax1];
			if (w1 >= 0 ) {
				w2 = subst_matrix[aa2][aax2];
				w = (w2 >= 0) ? w1+w2 : w1;

				if(w >= cutoff) {
					if(!pair_families[i]) {
						pair_families[i] = (long*)malloc(MAX_PAIR_FAMILY * sizeof(long));
						pair_families[i][0] = 0;
					}
					idx = pair_families[i][0];
					pair_families[i][++idx] = j;
					pair_families[i][0] = idx;
					assert(idx < MAX_PAIR_FAMILY);
				}
			}
		}
	}
	return SUCCESS;
}

//------------------------------------------------------------
void pair_primer_families_release(kernel_compute_data* _data)
{
	unsigned i, count = sizeof(_data->pair_primer_data.pair_families) / sizeof(_data->pair_primer_data.pair_families[0]);

	assert(_data);
	assert(_data->params->method == kernel_method_pair_primers);

	for(i = 0; i < count; i++) {
		if(_data->pair_primer_data.pair_families[i]) {
			free(_data->pair_primer_data.pair_families[i]);
			_data->pair_primer_data.pair_families[i] = NULL;
		}
	}
}

//------------------------------------------------------------
int init_compute_pair_primer_kernels(sequence _seq1)
{
	char* aas = _seq1.aas;
	unsigned l = _seq1.length - 1, idx, i, j, count, related_idx;
	long_pointer family;
	long_pointer* pair_families = (long_pointer*)G_kernel_compute_data.pair_primer_data.pair_families;
	long_pointer* pair_starts = (long_pointer*)G_kernel_compute_data.pair_primer_data.pair_starts;
	memset(pair_starts, 0, j = sizeof(G_kernel_compute_data.pair_primer_data.pair_starts));
	
	for(i = 0; i < l; i++) {
		idx = aas[i] * 20 + aas[i + 1];
		family = pair_families[idx];
		count = family ? *family : 0;
		while(count) {
			related_idx = family[count--];
			if(!pair_starts[related_idx]) {
				pair_starts[related_idx] = (long*)malloc(MAX_PAIR_FAMILY * sizeof(long));
				assert(pair_starts[related_idx]);
				pair_starts[related_idx][0] = 0;
			}
			j = pair_starts[related_idx][0];
			pair_starts[related_idx][++j] = i;
			pair_starts[related_idx][0] = j;
			assert(j < MAX_PAIR_FAMILY);
		}
	}
	return SUCCESS;
}

//------------------------------------------------------------
void pair_primer_kernels_release(kernel_compute_data* _data)
{
	unsigned i, count = sizeof(_data->pair_primer_data.pair_starts) / sizeof(_data->pair_primer_data.pair_starts[0]);

	assert(_data);
	assert(_data->params->method == kernel_method_pair_primers);

	for(i = 0; i < count; i++) {
		if(_data->pair_primer_data.pair_starts[i]) {
			free(_data->pair_primer_data.pair_starts[i]);
			_data->pair_primer_data.pair_starts[i] = NULL;
		}
	}
}

//------------------------------------------------------------
int compute_pair_primer_kernels(
	sequence* _seq1,
	sequence* _seq2, 
	kernel_params* _params
)
{
	char *aas = _seq2->aas;
	unsigned i, l = _seq2->length, j, mask_size, idx, pair_count;
	long_pointer* pair_starts = (long_pointer*)G_kernel_compute_data.pair_primer_data.pair_starts;
	long_pointer  related_pairs;

	// initialization
	if (G_kernel_compute_data.mask_matrix) free(G_kernel_compute_data.mask_matrix);
	G_kernel_compute_data.mask_matrix = (char*)malloc(mask_size = _seq1->length * _seq2->length * sizeof(char));
	memset(G_kernel_compute_data.mask_matrix, 0, mask_size);
	G_kernel_compute_data.seq1 = _seq1;
	G_kernel_compute_data.seq2 = _seq2;
	G_kernel_compute_data.params = _params;
	G_kernel_compute_data.kernel_count = 0;
	G_kernel_compute_data.coverage = 0;
	G_kernel_compute_data.strong_coverage = 0;
	G_kernel_compute_data.total_length = 0;
	G_kernel_compute_data.used_length = 0;
		
	i = l - 1;
	while(i--) {
		idx = aas[i] * 20 + aas[i + 1];
		related_pairs = pair_starts[idx];
		pair_count = related_pairs ? *related_pairs : 0;
		while(pair_count) {
			j = related_pairs[pair_count--];
			if(!G_kernel_compute_data.mask_matrix[j * l + i])
				check_hsp_arround_match(j, i);
		}
	}

	G_kernel_compute_data.seq1 = G_kernel_compute_data.seq2 = NULL;

	//debug_dump_kernels(G_debug_dump, &G_kernel_compute_data);
	free(G_kernel_compute_data.mask_matrix);
	G_kernel_compute_data.mask_matrix = NULL;
	
	return G_kernel_compute_data.kernel_count;
}

//--------------------------------------------------------------------------------
void debug_dump_kernels(FILE* _file, kernel_compute_data* _data )
{
	int i;
	for(i = 0; i < _data->kernel_count; i++)
		fprintf(_file, "(%d,%d)\t%d\t%.1f\n", _data->kernels[i].x, _data->kernels[i].y, _data->kernels[i].length, _data->kernels[i].score);
}

int debug_print_kernels(char* _file_name, kernel* _kernels, unsigned _kernel_count, char* _seqname1, char* _seqname2, unsigned _len1, unsigned _len2, float _id){

	FILE* f;
	int i;

	f = fopen(_file_name, "at");
	if (f == NULL) {
		printf("Cannot open kernels file: %s", _file_name);
		return 0;
	}

	fprintf(f,"name1\tname2\tLen1\tlen2\tID\t");
	fprintf(f,"nom\tX\tY\tLen\tScor\t\n");
	for (i=0; i<_kernel_count; i++){
		fprintf(f, "%s\t%s\t%d\t%d\t%f\t", _seqname1, _seqname2,  _len1, _len2, _id);
		fprintf(f,"%d\t%d\t%d\t%d\t%f\t\n", i,_kernels[i].x,_kernels[i].y, _kernels[i].length, _kernels[i].score);
	}
	fclose(f);
	return 1;
}

// kernel sorts
int kernel_weight_compare(const void * arg1, const void * arg2)
{
	const kernel* krn1 = (const kernel*)arg1;
	const kernel* krn2 = (const kernel*)arg2;
	if(krn1->score > krn2->score) return -1;
	else if (krn1->score < krn2->score) return 1;
	return 0;
}

//------------------------------------------------------------------------------------
unsigned reget_kernels(kernel* _kernels, unsigned _count, unsigned _required)
{
	qsort(_kernels, _count, sizeof(kernel), kernel_weight_compare);
	return __min(_count, _required);
}

//------------------------------------------------------------------------------------
unsigned reget_point_count(kernel* _kernels, unsigned _count)
{
	int i;
	unsigned point_count = 0;
	for (i=0; i<_count; i++) {
		point_count += _kernels[i].length;
	}
	return point_count;
}

//---------------------------------------------------
int kernel_compare(const void * arg1, const void * arg2)
{
	const kernel* krn1 = (const kernel*)arg1;
	const kernel* krn2 = (const kernel*)arg2;
	if(krn1->y < krn2->y) return -1;
	else if (krn1->y > krn2->y) return 1;
	return krn1->x < krn2->x ? -1 : 1;
}

//-----------------------------------------------------
void mysort(kernel* _kernels, unsigned _count)
{
	qsort(_kernels, _count, sizeof(kernel), kernel_compare);
}

//-----------------------------------------------------
unsigned get_kern_count(kernel* kernels, unsigned kernel_count, score_type weight){
  unsigned count =0, i;
  for (i=0; i<kernel_count; i++) {
	  if (kernels[i].score >= weight) count++;
  }

  return count;
}


//-----------------------------------------------------
// debug print matrixes
void print_pair_primer_families(char* _file_name){

	int i,j, nom;
	FILE* f;
	long_pointer* pair_families = (long_pointer*)G_kernel_compute_data.pair_primer_data.pair_families;

	f = fopen(_file_name, "wt");
	if (f == NULL) {
		printf("Cannot open kernels file: %s", _file_name);
		return;
	}
    fprintf(f, "pair_primer_families\n");
	fprintf(f, "indx\tnom\tfrends\n\t");

	for (i=0; i<400; i++){
		fprintf(f, "%d\t", i);
		nom = pair_families[i]? pair_families[i][0] : 0;
		fprintf(f, "%d\t", nom);

		for (j=1; j<nom+1; j++) fprintf(f, "%d\t", pair_families[i][j]);
		fprintf(f,"\n");
	}
	fclose(f);
	fprintf(f,"\n");
}

void print_pair_starts(char* _file_name){
	int i,j, nom;
	FILE* f;
	long_pointer* pair_starts = (long_pointer*)G_kernel_compute_data.pair_primer_data.pair_starts;

	f = fopen(_file_name, "wt");
	if (f == NULL) {
		printf("Cannot open kernels file: %s", _file_name);
		return;
	}
    fprintf(f, "pair_primer_families\n");
	fprintf(f, "indx\tnom\tstarts\n\t");

	for (i=0; i<400; i++){
		fprintf(f, "%d\t", i);
		nom = pair_starts[i]? pair_starts[i][0] : 0;
		fprintf(f, "%d\t", nom);

		for (j=1; j<nom+1; j++) fprintf(f, "%d\t", pair_starts[i][j]);
		fprintf(f,"\n");
	}
	fclose(f);
}

