
///////////////////////////////////////////////////////////////
//
//	File:		sparse.c
//	Created:	20.12.2001
//	Purpose:	Sparse dynamic programming in sequence alignment
//
//	Comments:	
//	Authors:	Bogopolsky Gennady, EIMB (c) 2001
//
///////////////////////////////////////////////////////////////

#include "sparse.h"

FILE*			G_debug_dump;
hexograph		G_hexograph;
sparse_alignment_data G_sparse_data;

hexograph_node*	G_node_pool;

char compute_hexograph_node_start(hexograph_node* _node);
char compute_hexograph_node_inner(hexograph_node* _node);
void hexograph_path_improve(coord_type _i1, coord_type _j1,
							coord_type _i2, coord_type _j2,
							char* _seq1, char* _seq2, int* _path);
void hexograph_sw__doublegap_improve(coord_type _i1, coord_type _j1,
						coord_type _i2, coord_type _j2,
						char _free_begins, char _free_ends);

void hexograph_cut_negative();

void hexograph_sw_improve(coord_type _i1, coord_type _j1,
						coord_type _i2, coord_type _j2,
						char _free_begins, char _free_ends);
void debug_dump_hexograph(FILE* _file, hexograph_node* _node);

//---------------------------------------------------
/*int kernel_compare(const void * arg1, const void * arg2)
{
	const kernel* krn1 = (const kernel*)arg1;
	const kernel* krn2 = (const kernel*)arg2;
	if(krn1->y < krn2->y) return -1;
	else if (krn1->y > krn2->y) return 1;
	return krn1->x < krn2->x ? -1 : 1;
}*/

//---------------------------------------------------
kernel_list_node* kernel_list_insert(
	kernel_list* _list, kernel* _krn, kernel_list_node* _before)
{

	kernel_list_node* new_node;

	assert(_list);
	assert(_krn);

	if(_list->head == NULL) {
		assert(_before == NULL);
		_list->head = (kernel_list_node*)malloc(sizeof(kernel_list_node));
		_list->head->data = *_krn;
		_list->head->start = 0;
		_list->head->prev = _list->head->next = NULL;
		_list->tail = _list->head;
		return _list->head;
	}
	if(_before == NULL) { // insert at end
		assert(_list->head != NULL && _list->tail != NULL);
		new_node = (kernel_list_node*)malloc(sizeof(kernel_list_node));
		new_node->data = *_krn;
		new_node->start = 0;
		// set links
		new_node->next = NULL;
		new_node->prev = _list->tail;
		_list->tail->next = new_node;
		_list->tail = new_node;
		return new_node;
	}

	new_node = (kernel_list_node*)malloc(sizeof(kernel_list_node));
	new_node->data = *_krn;
	new_node->start = 0;
	// set links
	new_node->next = _before;
	new_node->prev = _before->prev;
	_before->prev = new_node;
	if(new_node->prev) new_node->prev->next = new_node;
	else _list->head = new_node;
	return new_node;
}

//---------------------------------------------------
kernel_list_node* kernel_list_delete(kernel_list* _list, kernel_list_node* _node)
{
	// relink
	kernel_list_node* ret = _node->next;
	
	assert(_node);

	if(_node->prev) _node->prev->next = _node->next;
	if(_node->next) _node->next->prev = _node->prev;
	if(_list->head == _node) _list->head = _node->next;
	if(_list->tail == _node) _list->tail = _node->prev;
	// release
	free(_node);
	return ret;
}

//---------------------------------------------------
coord_type hexograph_diag_diff(hexograph_node* _p, hexograph_node* _q)
{
	return ABS(_p->diag - _q->diag);
}

/*void mysort(kernel* _kernels, unsigned _count)
{
	qsort(_kernels, _count, sizeof(kernel), kernel_compare);
}*/

//---------------------------------------------------
int hexograph_build(kernel* _kernels, unsigned _count, unsigned _point_count, sequence* _seq1, sequence* _seq2)
{
	int i, j, m = _seq1->length, n = _seq2->length, *path = G_sparse_data.path, prev_diag = 0, prev_kernel = 0;

	kernel* current_kernel = _kernels;
	kernel_list				open_kernels = { 0 };
	kernel_list_node*		open_krn_iterator = NULL;
	hexograph_node*  prev_node = 0, *current_node, *max_node = NULL; //  
	hexograph_node   *prev_top_nodes[1024];  //  = (ptr_type*)malloc((_len1 ) * sizeof(ptr_type));
	hexograph_node   *diag_nodes[2048]; // = (ptr_type*)malloc((_len1 + _len2) * sizeof(ptr_type));
	hexograph_node** prev_diag_nodes = (hexograph_node**)diag_nodes;
	score_type		max_score = 0, gop, gep;

	G_sparse_data.seq1 = _seq1;
	G_sparse_data.seq2 = _seq2;
	G_sparse_data.owners_tail = G_sparse_data.owners;

	//    x   y
	mysort(_kernels, _count);
	// qsort(_kernels, _count, sizeof(kernel), kernel_compare);

	//  
	G_hexograph.pool_size = _point_count + 1;
	 G_hexograph.node_pool = (hexograph_node*)malloc(G_hexograph.pool_size * sizeof(hexograph_node));
	assert(G_hexograph.node_pool);
	memset(G_hexograph.node_pool, 0, G_hexograph.pool_size * sizeof(hexograph_node));
	current_node = G_hexograph.node_pool;
	
	memset(prev_top_nodes, 0, m * sizeof(hexograph_node*));
	memset(prev_diag_nodes, 0, (m + n) * sizeof(hexograph_node*));
	//        ,  
	prev_diag_nodes += n;

	//       !
	//assert(_kernels);
	//assert(_count > 0);

	_kernels += _count;
	//   
	for(j = 0; j < n || open_kernels.head != NULL; j++) {

		// j -  

		open_krn_iterator = open_kernels.head;
		//       
		while(current_kernel != _kernels && current_kernel->y == j) {
			while(open_krn_iterator && current_kernel->x > open_krn_iterator->data.x)
				open_krn_iterator = open_krn_iterator->next;
			//   open_krn_iterator
			kernel_list_insert(&open_kernels, current_kernel, open_krn_iterator);
			current_kernel ++ ;
		}

		prev_node = 0;
		open_krn_iterator = open_kernels.head;
		while(open_krn_iterator) {
			
			i = open_krn_iterator->data.x;

			// assert(G_hexograph.pool_ptr < G_hexograph.pool_size);

			current_node->point_of_kernel = open_krn_iterator->data.id;

			current_node->left = prev_node;
			prev_node = current_node;

			current_node->prev = prev_diag_nodes[i - j];
			prev_diag_nodes[i - j] = current_node;

			current_node->top = prev_top_nodes[i];
			prev_top_nodes[i] = current_node;

			current_node->x_coord = i; current_node->y_coord = j; current_node->diag = i - j;

			if(open_krn_iterator->start) compute_hexograph_node_inner(current_node); 
			else compute_hexograph_node_start(current_node); 

			//debug_dump_hexograph(G_debug_dump, current_node);

			if(--open_krn_iterator->data.length) {
				open_krn_iterator->data.x++;
				open_krn_iterator->data.y++;
				open_krn_iterator->start = 1; // 
				open_krn_iterator = open_krn_iterator->next;
			} else {
				//  
				open_krn_iterator = kernel_list_delete(&open_kernels, open_krn_iterator);

				if(current_node->strong) {
					*(G_sparse_data.owners_tail) = current_node;
					G_sparse_data.owners_tail ++ ;
					if(current_node->best_score >= max_score) {
						max_score = current_node->best_score;
						max_node = current_node;
					}
				}
			}

			current_node ++ ;

			if(!open_krn_iterator) break;
		}
	}

	G_hexograph.pool_tail = current_node;

	if(!max_node) {
		G_sparse_data.pure_sw ++ ;
		if(G_sparse_data.improve) hexograph_sw__doublegap_improve(-1, -1, m, n, 1, 1);
								//hexograph_sw_improve(-1, -1, m, n, 1, 1);
		return 0;
	}

	//  
	
	//  -1 
	for(i = 0; i < m; i++) path[i] = -1;

	gop = G_sparse_data.GOP;
	gep = G_sparse_data.GEP;
	//G_sparse_data.GOP = G_sparse_data.KCP;
	//G_sparse_data.GEP = G_sparse_data.KEP;

	if(G_sparse_data.improve) //hexograph_sw__doublegap_improve(max_node->x_coord, max_node->y_coord, m, n, 1, 1);
					 hexograph_sw__doublegap_improve(max_node->x_coord, max_node->y_coord, m, n, 0, 1);
					// hexograph_sw_improve(max_node->x_coord, max_node->y_coord, m, n, 0, 1);

	//G_sparse_data.GOP = gop;
	//G_sparse_data.GEP = gep;

	while(1) {
		//    
		i = max_node->x_coord;
		j = max_node->y_coord;
		path[i] = j;
		prev_kernel = max_node->point_of_kernel;
		// prev_diag = max_node->diag;
		max_node = max_node->best_link;
		if(!max_node) break;
		if(/*prev_diag == max_node->diag &&*/ prev_kernel == max_node->point_of_kernel) {
			while(i > max_node->x_coord && j > max_node->y_coord) path[i--] = j--;
		} else {
			if(i == max_node->x_coord + 1 || j == max_node->y_coord + 1)
				while(--i > max_node->x_coord) path[i] = -1;
			else {
				if(G_sparse_data.improve) //hexograph_sw__doublegap_improve(max_node->x_coord, max_node->y_coord, i, j, 1, 1);
					hexograph_sw__doublegap_improve(max_node->x_coord, max_node->y_coord, i, j, 0, 0);
				//if(G_sparse_data.improve) hexograph_sw_improve(max_node->x_coord, max_node->y_coord, i, j, 0, 0);
			}
		}
	}

	G_sparse_data.GOP = G_sparse_data.KCP;
	G_sparse_data.GEP = G_sparse_data.KEP;

	if(G_sparse_data.improve) //hexograph_sw__doublegap_improve(-1, -1, i, j, 1, 1);
							 hexograph_sw__doublegap_improve(-1, -1, i, j, 1, 0);
							//hexograph_sw_improve(-1, -1, i, j, 1, 0);

	G_sparse_data.GOP = gop;
	G_sparse_data.GEP = gep;

	//hexograph_cut_negative();

	G_sparse_data.max_score = max_score;
	//assert(open_kernels.head == NULL);
	//   -  
	//	free(prev_top_nodes);
	//	free(prev_diag_nodes - _len2);

	return SUCCESS;
}

//--------------------------------------------------
void hexograph_cut_negative()
{
	int				i = 0, m = G_sparse_data.seq1->length;
	score_type		w_max = 0;
	char*			seq1 = G_sparse_data.seq1->aas, *seq2 = G_sparse_data.seq2->aas; 
	int*			path = G_sparse_data.path;

	assert(seq1);
	assert(seq2);

	// 
	while(path[i] == -1) i++;

	do {
		w_max += subst_matrix[seq1[i]][seq2[path[i]]];
		i++;
	} while(path[i] != -1 && path[i] == path[i-1] - 1);
	if(w_max < 0) //    
		while(i--) path[i] = -1;

	// 
	i = m - 1;
	while(path[i] == -1) i--;

	w_max = 0;
	do {
		w_max += subst_matrix[seq1[i]][seq2[path[i]]];
	 	i--;
	} while(path[i] != -1 && path[i] == path[i+1] + 1);
	if(w_max < 0) //    
		while(i < m) path[i++] = -1;

}

//--------------------------------------------------
void hexograph_path_improve(coord_type _i1, coord_type _j1,
							coord_type _i2, coord_type _j2,
							char* _seq1, char* _seq2, int* _path
						//	,char free_begin_, char free_end_
							)
{
	score_type sums[MAX_LEN], sum_max = -1000, sum;
	coord_type i, j, k = 0, k_max = 0;
	if(_i2 - _i1 < 2 || _j2 - _j1 < 2) return; // there is nothing to improve

	i = _i1 + 1, j = _j1 + 1;
	sums[0] = sum = 0;


	while(i < _i2 && j < _j2) {
		sum = sum + subst_matrix[_seq1[i++]][_seq2[j++]];
		sums[++k] = sum;
		if (sum > sum_max) {
			sum_max = sum;
			k_max = k;
		}
	}

	i = _i2 - 1, j = _j2 - 1;
	sum = 0;
	sum_max = sums[k];
	k_max = k--;
	while(i > _i1 && j > _j1) {
		sum = sum + subst_matrix[_seq1[i--]][_seq2[j--]];
		k--;
	}
	assert(k == -1);
	// optimal k is now found
	if (sum_max<=0) return; // skip negative impruve

	k = k_max;
	while(k--) _path[++_i1] = ++_j1;
	
	if(_i2 - _i1 > _j2 - _j1) {
		_i1 += (_i2 - _i1) - (_j2 - _j1);
	} else {
		_j1 += (_j2 - _j1) - (_i2 - _i1);
	}

	while(_i1 < _i2 && _j1 < _j2) {
		_path[_i1++] = _j1++;
	}
}

//--------------------------------------------------
// lokal path impruve
void hexograph_path_improve_lokal(coord_type _i1, coord_type _j1,
							coord_type _i2, coord_type _j2,
							char* _seq1, char* _seq2, int* _path
							,char free_begin_, char free_end_
							)
{
	score_type sums[MAX_LEN], sum_max = -1000, sum, sum_res;
	coord_type i, j, k = 0, k_max = 0;
	if(_i2 - _i1 < 2 || _j2 - _j1 < 2) return; // there is nothing to improve

	i = _i1 + 1, j = _j1 + 1;
	sums[0] = sum = 0;

	while(i < _i2 && j < _j2) {
		sum = sum + subst_matrix[_seq1[i++]][_seq2[j++]];
		sums[++k] = (!free_begin_) ? sum: 0;
	}

	i = _i2 - 1, j = _j2 - 1;
	sum = 0;
	sum_max = sums[k];
	k_max = k--;
	while(i > _i1 && j > _j1) {
		sum = sum + subst_matrix[_seq1[i--]][_seq2[j--]];
		sum_res = (!free_end_) ? sum: 0;
		if(sum_res + sums[k] > sum_max) {
			sum_max = sum_res + sums[k];
			k_max = k;
		}
		k--;
	}
	assert(k == -1);
	// optimal k is now found
	if (sum_max<=0) return; // skip negative impruve

	k = k_max;
	if (!free_begin_) while(k--) _path[++_i1] = ++_j1;
	else { _i1+=k; _j1+=k;  k=0;} 
	
	if(_i2 - _i1 > _j2 - _j1) {
		_i1 += (_i2 - _i1) - (_j2 - _j1);
	} else {
		_j1 += (_j2 - _j1) - (_i2 - _i1);
	}

	if (!free_end_) while(_i1 < _i2 && _j1 < _j2) _path[_i1++] = _j1++;
}

//--------------------------------------------------
char compute_hexograph_node_start(hexograph_node* _node)
{
	hexograph_node** iterator = G_sparse_data.owners;
	hexograph_node*	 max_owner = NULL, *left, *top, *jump, *owner;
	coord_type diag = _node->diag, diff;
	score_type max_owner_weight = 0;
	score_type w_left = 0, w_top = 0, w_diag, w, w_jump;
	score_type kcp = G_sparse_data.KCP, gep = G_sparse_data.KEP;
	char strong = 1;

	assert(_node);

	//  
	while(iterator != G_sparse_data.owners_tail) {
		diff = ABS((*iterator)->diag - diag);
		if((*iterator)->x_coord < _node->x_coord && (*iterator)->y_coord < _node->y_coord) {
			if((w = (*iterator)->jump_score - diff * gep) >= max_owner_weight) {
				max_owner_weight = w;
				max_owner = *iterator;
			}
		}
		iterator ++;
	}

	// 

	//      
	owner = left = _node->left;
	while(owner) {
		if(owner->strong && owner->y_coord < _node->y_coord) {
			diff = ABS(diag - owner->diag);
			if((w = owner->jump_score - diff * gep) > max_owner_weight) {
				max_owner_weight = w;
				max_owner = owner;
				break;
			}
		}
		owner = owner->best_link;
	}
	owner = top = _node->top;
	while(owner) {
		if(owner->strong && owner->x_coord < _node->x_coord) {
			diff = ABS(diag - owner->diag);
			if((w = owner->jump_score - diff * gep) > max_owner_weight) {
				max_owner_weight = w;
				max_owner = owner;
				break;
			}
		}
		owner = owner->best_link;
	}

	if(max_owner) {
		if(max_owner_weight < 0) {
			max_owner_weight = 0;
			max_owner = 0;
		}
	}
	
	w_diag = max_owner_weight + subst_matrix[G_sparse_data.seq1->aas[_node->x_coord]][G_sparse_data.seq2->aas[_node->y_coord]];

	if(left) {
		w_left = left->jump_score - gep * ABS(left->diag - diag);
		left = left->jump_link;
	};
	if(top) {
		w_top = top->jump_score - gep * ABS(top->diag - diag);
		top = top->jump_link;
	};

	if(w_top > w_left) { w_jump = w_top; jump = top; }
	else { w_jump = w_left; jump = left; }
	
	if(w_diag > w_jump) {
		_node->best_score = w_diag;
		_node->best_link = max_owner;

		if(w_diag - kcp >= w_jump) { 
			_node->jump_score = w_diag - kcp;
			_node->jump_link = max_owner;
		} else { 
			_node->jump_score = w_jump;
			_node->jump_link = jump ? jump->jump_link : NULL;
		};
	} else {
		_node->best_score = _node->jump_score = w_jump;
		_node->best_link = _node->jump_link = jump;
		strong = 0;
	}
	return (_node->strong = strong);
}

//--------------------------------------------------
char compute_hexograph_node_inner(hexograph_node* _node)
{
	score_type kcp = G_sparse_data.KCP, gep = G_sparse_data.KEP;
	score_type w_left = 0, w_top = 0, w_jump, w;
	hexograph_node* left = _node->left, *top = _node->top, *jump = NULL, *prev = _node->prev;
	score_type w_diag = prev->best_score + subst_matrix[G_sparse_data.seq1->aas[_node->x_coord]][G_sparse_data.seq2->aas[_node->y_coord]];
	coord_type diag = _node->diag;
	char strong = 1;

	assert(_node);

	if(left) {
		w_left = left->jump_score - gep * ABS(left->diag - diag);
		left = left->strong ? left : left->jump_link;
	};
	if(top) {
		w_top = top->jump_score - gep * ABS(top->diag - diag);
		top = top->strong ? top : top->jump_link;
	};
	if(w_left > w_top) { w_jump = w_left; jump = left; }
	else { w_jump = w_top; jump = top; };
	if(prev->jump_score > w_jump) { w_jump = prev->jump_score; jump = prev->jump_link;};
	
	if(w_diag >= w_jump) {
		_node->best_score = w_diag;
		_node->best_link = prev->strong ? prev : prev->best_link;

		if((w = w_diag - kcp) >= w_jump) {
			_node->jump_score = w;
			_node->jump_link = _node /*prev->strong ? prev : prev->best_link*/; //   
		} else {
			_node->jump_score = w_jump;
			_node->jump_link = jump ? jump->jump_link : NULL;
		}

		if(_node->jump_score < prev->jump_score) {
			_node->jump_score = prev->jump_score;
			_node->jump_link = prev->jump_link;
			strong = 0;
		}
	} else {

		_node->best_score = _node->jump_score = w_jump;
		_node->best_link = _node->jump_link = jump ? jump->jump_link : NULL;
		
		if(_node->jump_score < prev->jump_score) {
			_node->jump_score = prev->jump_score;
			_node->jump_link = prev->jump_link;
		}
		strong = 0;
	}
	return (_node->strong = strong);
}

//--------------------------------------------------
void hexograph_release(hexograph* _hexograph)
{
	if (_hexograph->node_pool) free(_hexograph->node_pool);
	_hexograph->node_pool = 0;
	_hexograph->pool_size = 0;
	_hexograph->total_count = 0;
	_hexograph->pool_tail = _hexograph->node_pool;
}

//--------------------------------------------------
void hexograph_dbg_print(FILE* _f)
{
	int i = 0;
	hexograph_node* node = G_hexograph.node_pool;
	fprintf(_f, "\nhexograph:\n");
	while(node != G_hexograph.pool_tail) {
		fprintf(_f, "%d\t%d\t%d\t%d\n", i ++ ,
			node->left ? node->left - G_hexograph.node_pool : -1, 
			node->prev ? node->prev - G_hexograph.node_pool : -1,
			node->top ? node->top - G_hexograph.node_pool : -1
		);
		node++;
	}
}

//--------------------------------------------------
void hexograph_sw_improve(coord_type _i1, coord_type _j1,
						coord_type _i2, coord_type _j2,
						char _free_begins, char _free_ends
)
{
	int				m = _i2 - (++_i1), n = _j2 - (++_j1);
	int				i, j, i_max = -1, j_max = -1;
	score_type		GOP = G_sparse_data.GOP, GEP = G_sparse_data.GEP, w_max;
	sw_link			*sw_links;
	sw_cell			*current_line, *prev_line, *p = NULL;
	char			dir, open_flag;
	char*			seq1 = G_sparse_data.seq1->aas, *seq2 = G_sparse_data.seq2->aas; 
	int*			path = G_sparse_data.path;

	assert(seq1);
	assert(seq2);

	if(m <= 0 || n <= 0) return;

	// general memory allocation for NW-graph
	current_line = (sw_cell*)malloc((m + 1) * sizeof(sw_cell));
	assert(current_line);
	prev_line = (sw_cell*)malloc((m + 1) * sizeof(sw_cell));
	assert(prev_line);

	sw_links = (sw_link*)malloc(m * n * sizeof(sw_link));
	assert(sw_links);
	memset(sw_links, 0, m * n * sizeof(sw_link));
	
	if(_free_begins) {
		memset(prev_line, 0, (m + 1) * sizeof(sw_cell));
		memset(current_line, 0, sizeof(sw_cell));
	} else {
		prev_line[0].diag = 0;
		for(i = 1; i <= m; i++) {
			prev_line[i].top = -10000;
			prev_line[i].diag = - GOP - GEP * i;
		}
	}
	w_max = (score_type)- 10000 * (1-_free_ends);

	current_line++;
	prev_line++;

	//   
	for(j = 0; j < n; j++)
	{
		if(!_free_begins) {
			current_line[-1].left = -10000;
			current_line[-1].diag = - GOP - GEP * (j + 1);
		}

		//  
		for(i = 0; i < m; i++) {

			current_line[i].diag = prev_line[i-1].diag + subst_matrix[seq1[_i1 + i]][seq2[_j1 + j]];

			// 5-comparison goes here
			dir = 0; open_flag = 0;

			if(current_line[i-1].left < current_line[i-1].diag - GOP) {
				open_flag |= DIR_LEFTOPEN;
				current_line[i].left = current_line[i-1].diag - GOP - GEP;
			} else 
				current_line[i].left = current_line[i-1].left - GEP;

			if(prev_line[i].top < prev_line[i].diag - GOP) {
				open_flag |= DIR_TOPOPEN;
				current_line[i].top = prev_line[i].diag - GOP - GEP;
			} else 
				current_line[i].top = prev_line[i].top - GEP;

			if(current_line[i].top > current_line[i].diag) {
				if(current_line[i].top > current_line[i].left) {
					current_line[i].diag = current_line[i].top;
					dir = DIR_TOP;
				} else {
					current_line[i].diag = current_line[i].left;
					dir = DIR_LEFT;
				}
			} else if(current_line[i].left > current_line[i].diag) {
				current_line[i].diag = current_line[i].left;
				dir = DIR_LEFT;
			} else {
				dir = DIR_DIAG;
			}

			if(_free_begins) {
				if(current_line[i].diag < 0) {
					current_line[i].diag = 0;
					dir = DIR_TERMINATOR;
				}
			}
			//if(_free_ends) {
				if(current_line[i].diag >= w_max) {
					w_max = current_line[i].diag;
					i_max = i; j_max = j;
				}
			//}

			sw_links[j * m + i].link = dir;
			sw_links[j * m + i].open_flag = open_flag;
		}
		// swap prev/current
		p = prev_line;
		prev_line = current_line;
		current_line = p;
	}

	if(w_max > 0) { //  imptovement !!! 
					//   ,  
		//  ,    
		if(_free_ends) {
			for(i = m - 1; i > i_max; path[_i1 + i--] = -1);
			i = i_max; j = j_max;
		} else {
			i = m - 1; j = n - 1;
		}
		while(i >= 0 && j >= 0) {
			switch(sw_links[j * m + i].link) {
			case DIR_DIAG:
				path[_i1 + i--] = _j1 + j--;
				break;
			case DIR_TOP :
				do {
					assert(j >= 0);
					path[_i1 + i] = _j1 + j; 
				} while (!(sw_links[(j--) * m + i].open_flag & DIR_TOPOPEN));
				break;
			case DIR_LEFT :
				do {
					assert(i >= 0);
					path[_i1 + i] = -1;
				} while (!(sw_links[j * m + i--].open_flag & DIR_LEFTOPEN));
				break;
			default:
				assert((sw_links[j * m + i].link == DIR_TERMINATOR) && _free_begins);
				j = -1; // in order to exit while !!!
				break;
			}
		}

		if(_free_begins) {
			while(i >= 0) path[_i1 + i--] = -1;
		}
	}

	free(sw_links);
	free(--current_line);
	free(--prev_line);
}

//--------------------------------------------------------------------------------
void hexograph_sw__doublegap_improve(coord_type _i1, coord_type _j1,
						coord_type _i2, coord_type _j2,
						char _free_begins, char _free_ends
)
{
	int				m = _i2 - (++_i1), n = _j2 - (++_j1);
	int				i, j, i_max = -1, j_max = -1;
	score_type		GOP = G_sparse_data.GOP, GEP = G_sparse_data.GEP, w_max , jump_left = 0, jump_top = 0;
	sw_link			*sw_links;
	sw_cell			*current_line, *prev_line, *p = NULL;
	char			dir, open_flag;
	char*			seq1 = G_sparse_data.seq1->aas, *seq2 = G_sparse_data.seq2->aas; 
	int*			path = G_sparse_data.path;

	assert(seq1);
	assert(seq2);

	if(m <= 0 || n <= 0) return;

	// general memory allocation for NW-graph
	current_line = (sw_cell*)malloc((m + 1) * sizeof(sw_cell));
	assert(current_line);
	prev_line = (sw_cell*)malloc((m + 1) * sizeof(sw_cell));
	assert(prev_line);

	sw_links = (sw_link*)malloc(m * n * sizeof(sw_link));
	assert(sw_links);
	memset(sw_links, 0, m * n * sizeof(sw_link));
	
	if(_free_begins) {
		memset(prev_line, 0, (m + 1) * sizeof(sw_cell));
		memset(current_line, 0, sizeof(sw_cell));
	} else {
		prev_line[0].diag = 0;
		for(i = 1; i <= m; i++) {
			prev_line[i].jump = -10000;
			prev_line[i].diag = - GOP - GEP * i;
		}
	}
	w_max = 0;//(score_type)- 10000 * (1-_free_ends);

	current_line++;
	prev_line++;

	//   
	for(j = 0; j < n; j++)
	{
		//if(!_free_begins) {
			current_line[-1].jump = (score_type) -10000 * (1 - _free_begins);
			current_line[-1].diag = (- GOP - GEP * (j + 1)) * (1 - _free_begins);
		//}

		//  
		for(i = 0; i < m; i++) {

			current_line[i].diag = prev_line[i-1].diag + subst_matrix[seq1[_i1 + i]][seq2[_j1 + j]];

			// 5-comparison goes here
			dir = 0; open_flag = 0;

			if(current_line[i-1].jump < current_line[i-1].diag - GOP) {
				open_flag |= DIR_LEFTOPEN;
				jump_left = current_line[i-1].diag - GOP - GEP;
			} else 
				jump_left = current_line[i-1].jump - GEP;

			if(prev_line[i].jump < prev_line[i].diag - GOP) {
				open_flag |= DIR_TOPOPEN;
				jump_top = prev_line[i].diag - GOP - GEP;
			} else 
				jump_top = prev_line[i].jump - GEP;

			if(jump_top > jump_left) {
				current_line[i].jump = jump_top;
				dir = DIR_TOP;
			} else {
				current_line[i].jump = jump_left;
				dir = DIR_LEFT;
			}

			if(current_line[i].jump > current_line[i].diag) {
				current_line[i].diag = current_line[i].jump;
			} else {
				dir = DIR_DIAG;
			}

			if(_free_begins && (current_line[i].diag < 0)) {
					current_line[i].diag = 0;
					dir = DIR_TERMINATOR;
			}
			//if(_free_ends) {
				if(current_line[i].diag >= w_max) {
					w_max = current_line[i].diag;
					i_max = i; j_max = j;
				}
			//}

			sw_links[j * m + i].link = dir;
			sw_links[j * m + i].open_flag = open_flag;
		}
		// swap prev/current
		p = prev_line;
		prev_line = current_line;
		current_line = p;
	}

	//if(w_max < 0) { //  imptovement !!!  

		//  ,    
		if(_free_ends) {
			for(i = m - 1; i > i_max; path[_i1 + i--] = -1);
			i = i_max; j = j_max;
		} else {
			i = m - 1; j = n - 1;
		}
		while(i >= 0 && j >= 0) {
			switch(sw_links[j * m + i].link) {
			case DIR_DIAG:
				path[_i1 + i--] = _j1 + j--;
				break;
			case DIR_TOP :
				do {
					if(sw_links[j * m + i].link == DIR_LEFT) break;
					assert(j >= 0);
					path[_i1 + i] = _j1 + j; 
				} while (!(sw_links[(j--) * m + i].open_flag & DIR_TOPOPEN));
				break;
			case DIR_LEFT :
				do {
					if(sw_links[j * m + i].link == DIR_TOP) break;
					assert(i >= 0);
					path[_i1 + i] = -1;
				} while (!(sw_links[j * m + i--].open_flag & DIR_LEFTOPEN));
				break;
			default:
				assert((sw_links[j * m + i].link == DIR_TERMINATOR) && _free_begins);
				j = -1; // in order to exit while !!!
				break;
			}
		}

		if(_free_begins) {
			while(i >= 0) path[_i1 + i--] = -1;
		}
	//}

	free(sw_links);
	free(--current_line);
	free(--prev_line);
}

//--------------------------------------------------------------------------------
void debug_dump_hexograph(FILE* _file, hexograph_node* _node)
{
	fprintf(_file, "(%d,%d)\t%.1f\t%.1f\t->(%d,%d)\n", _node->x_coord, _node->y_coord, _node->best_score, _node->jump_score, _node->best_link?_node->best_link->x_coord:-1, _node->best_link?_node->best_link->y_coord:-1);
}
