
#ifndef _sparse_h_
#define _sparse_h_

///////////////////////////////////////////////////////////////
//
//	File:		sparse.h
//	Created:	20.12.2001
//	Purpose:	Sparse dynamic programming in sequence alignment
//
//	Comments:	
//	Authors:	Bogopolsky Gennady, EIMB (c) 2001
//
///////////////////////////////////////////////////////////////


#include "skoplib.h"


#define			MINIMAL			-32000

#define			ABS(a)			((a)>0?(a):(-(a)))

/////////////////////////////////////////////////////////////////////////////////
//	SparseKOP - Alignment as it has to be :-)
/////////////////////////////////////////////////////////////////////////////////


/*
	Accepted sequence orientation:

	           n,j
     ----------------------->
	|       *
	|  *     *
 m,i|   *          *
	|    *          *
	|                *
	V

	The path is growed from left to rigth from up to bottom (similar to Excel).
*/

/* memory allocation PAGE_SIZE */
#define			PAGE_SIZE		16384

/*------------------------------------------------------------
	 Data Structures
------------------------------------------------------------*/

/*	Hexograph - the list with 6 links:
	
              n
     ------------------->
    |
	|	* *
  m	|	 \| 
	|	*-X-*
	|	  |\
	|	  * *
	V

*/

typedef struct _hexograph_node {
	struct _hexograph_node*	left;		//    
	struct _hexograph_node*	right;		//  
	struct _hexograph_node*	top;
	struct _hexograph_node*	bottom;
	struct _hexograph_node*	prev;
	struct _hexograph_node*	next;

	coord_type	x_coord;	//      
	coord_type	y_coord;	//      
	coord_type	diag;		//  

	score_type	end_score;		//    . ,    (x_coord, y_coord)
							// c   , ..  .
	struct _hexograph_node*	end_link;		//     score (*) ???

	score_type	best_score; //    .    [0..x_coord] vs. [0..y_coord]
	struct _hexograph_node*	best_link;	//     score (*) ???

	score_type	jump_score;	// " " ,        
							//   
	struct _hexograph_node*	jump_link; // 
	
	score_type	negative_score;	//      ,   (x_coord, y_coord)
	struct _hexograph_node*	negative_link;	//      ,      (top)
	score_type	positive_score;	//      ,   (x_coord, y_coord)
	struct _hexograph_node*	positive_link;	//      ,      (left)

	// char		node_type;	//  : 0 - , 1 - , NB:   
	char		strong;		// true -  . false - 

	unsigned	point_of_kernel;	//       

} hexograph_node;

typedef struct _hexograph
{
	hexograph_node*		node_pool;
	hexograph_node*		pool_tail; // beyond the list
	unsigned			pool_size;
	unsigned			total_count; // overall count of nodes in pool
} hexograph;

typedef struct _sparse_alignment_data
{
	sequence*	seq1;
	sequence*	seq2;
	score_type	KCP;
	score_type	KEP;
	score_type	GOP;
	score_type	GEP;
	hexograph_node*		owners[POOL_SIZE];
	hexograph_node**	owners_tail;
	int			path[MAX_LEN];
	unsigned	pure_sw;
	score_type	max_score;
	char		improve;

} sparse_alignment_data;


typedef struct _kernel_list_node {
	kernel	data;
	char	start;
	struct _kernel_list_node* prev;
	struct _kernel_list_node* next;
} kernel_list_node;

typedef struct _kernel_list {
	kernel_list_node*	head;
	kernel_list_node*	tail;
} kernel_list;

extern hexograph G_hexograph;
extern sparse_alignment_data G_sparse_data;

int hexograph_build(kernel* _kernels, unsigned _count, unsigned _point_count, sequence* _seq1, sequence* _seq2);
void hexograph_release(hexograph* _hexograph);



#endif //_sparse_h_
