
///////////////////////////////////////////////////////////////
//
//	File:		kernels.c
//	Created:	20.12.2001
//	Purpose:	All about kernels and their's search methods
//
//	Comments:	
//	Authors:	Bogopolsky Gennady, EIMB (c) 2001
//
///////////////////////////////////////////////////////////////


#include "skoplib.h"

// Global definitions
sw_compute_data		G_sw_compute_data;

int compute_sw_alignment(
	sequence*	_seq1,
	sequence*	_seq2,
	sw_params*	_params
)
{
	unsigned	m = _seq1->length, n = _seq2->length;
	int			i, j, i_max = 0, j_max = 0;
	char		*seq1 = _seq1->aas, *seq2 = _seq2->aas;
	score_type  GOP = _params->GOP, GEP = _params->GEP, w_max = 0;
	sw_link		*sw_links;
	sw_cell		*current_line = G_sw_compute_data.line1, *prev_line = G_sw_compute_data.line2, *p = NULL;
	int*		path = G_sw_compute_data.path;
	char		dir, open_flag;

	assert(_seq1);
	assert(_seq2);
	assert(_params);

	// initialization
	G_sw_compute_data.seq1 = _seq1;
	G_sw_compute_data.seq2 = _seq2;

	// general memory allocation for NW-graph
	sw_links = (sw_link*)malloc(m * n * sizeof(sw_link));
	assert(sw_links);
	memset(sw_links, 0, m * n * sizeof(sw_link));

	// prepare
	memset(current_line, 0, (n + 1) * sizeof(sw_cell));
	memset(prev_line, 0, (n + 1) * sizeof(sw_cell));
	current_line++;
	prev_line++;

	//   
	for(i = 0; i < m; i++)
	{
		// swap prev/current
		p = prev_line;
		prev_line = current_line;
		current_line = p;

		//  
		for(j = 0; j < n; j++) {

			current_line[j].diag = prev_line[j-1].diag + subst_matrix[seq1[i]][seq2[j]];

			// 5-comparison goes here
			dir = 0; open_flag = 0;

			if(current_line[j-1].left < current_line[j-1].diag - GOP) {
				open_flag |= DIR_LEFTOPEN;
				current_line[j].left = current_line[j-1].diag - GOP - GEP;
			} else 
				current_line[j].left = current_line[j-1].left - GEP;

			if(prev_line[j].top < prev_line[j].diag - GOP) {
				open_flag |= DIR_TOPOPEN;
				current_line[j].top = prev_line[j].diag - GOP - GEP;
			} else 
				current_line[j].top = prev_line[j].top - GEP;

			if(current_line[j].top > current_line[j].diag) {
				if(current_line[j].top > current_line[j].left) {
					current_line[j].diag = current_line[j].top;
					dir = DIR_TOP;
				} else {
					current_line[j].diag = current_line[j].left;
					dir = DIR_LEFT;
				}
			} else if(current_line[j].left > current_line[j].diag) {
				current_line[j].diag = current_line[j].left;
				dir = DIR_LEFT;
			} else {
				dir = DIR_DIAG;
			}

			if(current_line[j].diag < 0) {
				dir = DIR_TERMINATOR;
				current_line[j].diag = 0;
			} else if(current_line[j].diag >= w_max) {
				w_max = current_line[j].diag;
				i_max = i;
				j_max = j;
			}

			sw_links[i * n + j].link = dir;
			sw_links[i * n + j].open_flag = open_flag;
		}
	}

	//  ,    
	for(m = m - 1; m > i_max; path[m--] = -1);

	while(i_max >= 0 && j_max >= 0 && sw_links[i_max * n + j_max].link != DIR_TERMINATOR) {
		path[i_max] = j_max;
		switch(sw_links[i_max * n + j_max].link) {
		case DIR_DIAG:
			i_max--; j_max--;
			break;
		case DIR_LEFT :
			do {
				assert(j_max >= 0);
				path[i_max] = j_max; 
			} while (!(sw_links[i_max * n + j_max--].open_flag & DIR_LEFTOPEN));
			break;
		case DIR_TOP :
			do {
				assert(i_max >= 0);
				path[i_max] = -1;
			} while (!(sw_links[(i_max--) * n + j_max].open_flag & DIR_TOPOPEN));
			break;
		default:
			assert(0);
			return 0;// E_UNEXPECTED;
		}
	}

	free(sw_links);

	while(i_max >= 0) { path[i_max--] = -1; }

	G_sw_compute_data.max_weight = w_max;

	return SUCCESS;
}
//------------------------------------------------------------------------------
int compute_sw_alignment_seqsubst(	// need activated seq_matrix
	sequence*	_seq1,					// with _seq1
	sequence*	_seq2,
	sw_params*	_params
)
{
	unsigned	n = _seq1->length, m = _seq2->length;
	int			i, j, i_max = 0, j_max = 0;
	char		*seq0 = _seq1->aas, *seq = _seq2->aas;
	score_type  GOP = _params->GOP, GEP = _params->GEP, w_max = 0;
	sw_link		*sw_links;
	sw_cell		*current_line = G_sw_compute_data.line1, *prev_line = G_sw_compute_data.line2, *p = NULL;
	int*		path = G_sw_compute_data.path, path2[MAX_LEN];
	char		dir, open_flag;
	score_type  *letter_subst;

	assert(_seq1);
	assert(_seq2);
	assert(_params);

	// initialization
	G_sw_compute_data.seq1 = _seq1;
	G_sw_compute_data.seq2 = _seq2;

	// general memory allocation for NW-graph
	sw_links = (sw_link*)malloc(m * n * sizeof(sw_link));
	assert(sw_links);
	memset(sw_links, 0, m * n * sizeof(sw_link));

	// prepare
	memset(current_line, 0, (n + 1) * sizeof(sw_cell));
	memset(prev_line, 0, (n + 1) * sizeof(sw_cell));
	current_line++;
	prev_line++;

	//   
	for(i = 0; i < m; i++)
	{
		// swap prev/current
		p = prev_line;
		prev_line = current_line;
		current_line = p;

		letter_subst = seq_matrix + (seq[i] * n);

		//  
		for(j = 0; j < n; j++) {

			current_line[j].diag = prev_line[j-1].diag + (*letter_subst); //subst_matrix[seq1[i]][seq2[j]];
			letter_subst ++;

			// 5-comparison goes here
			dir = 0; open_flag = 0;

			if(current_line[j-1].left < current_line[j-1].diag - GOP) {
				open_flag |= DIR_LEFTOPEN;
				current_line[j].left = current_line[j-1].diag - GOP - GEP;
			} else 
				current_line[j].left = current_line[j-1].left - GEP;

			if(prev_line[j].top < prev_line[j].diag - GOP) {
				open_flag |= DIR_TOPOPEN;
				current_line[j].top = prev_line[j].diag - GOP - GEP;
			} else 
				current_line[j].top = prev_line[j].top - GEP;

			if(current_line[j].top > current_line[j].diag) {
				if(current_line[j].top > current_line[j].left) {
					current_line[j].diag = current_line[j].top;
					dir = DIR_TOP;
				} else {
					current_line[j].diag = current_line[j].left;
					dir = DIR_LEFT;
				}
			} else if(current_line[j].left > current_line[j].diag) {
				current_line[j].diag = current_line[j].left;
				dir = DIR_LEFT;
			} else {
				dir = DIR_DIAG;
			}

			if(current_line[j].diag < 0) {
				dir = DIR_TERMINATOR;
				current_line[j].diag = 0;
			} else if(current_line[j].diag >= w_max) {
				w_max = current_line[j].diag;
				i_max = i;
				j_max = j;
			}

			sw_links[i * n + j].link = dir;
			sw_links[i * n + j].open_flag = open_flag;
		}
	}
	//  ,    
	for(m = m - 1; m > i_max; path2[m--] = -1);

	while(i_max >= 0 && j_max >= 0 && sw_links[i_max * n + j_max].link != DIR_TERMINATOR) {
		path2[i_max] = j_max;
		switch(sw_links[i_max * n + j_max].link) {
		case DIR_DIAG:
			i_max--; j_max--;
			break;
		case DIR_LEFT :
			do {
				assert(j_max >= 0);
				path2[i_max] = j_max; 
			} while (!(sw_links[i_max * n + j_max--].open_flag & DIR_LEFTOPEN));
			break;
		case DIR_TOP :
			do {
				assert(i_max >= 0);
				path2[i_max] = -1;
			} while (!(sw_links[(i_max--) * n + j_max].open_flag & DIR_TOPOPEN));
			break;
		default:
			assert(0);
			return 0;// E_UNEXPECTED;
		}
	}

	free(sw_links);

	while(i_max >= 0) { path2[i_max--] = -1; }



	G_sw_compute_data.max_weight = w_max;

	return SUCCESS;
}

//------------------------------------------------------------------------------
int compute_sw_alignment_cut(
	sequence*	_seq1,
	sequence*	_seq2,
	sw_params*	_params
)
{
	unsigned	m = _seq1->length, n = _seq2->length;
	int			i, j, i_max = 0, j_max = 0;
	char		*seq1 = _seq1->aas, *seq2 = _seq2->aas;
	score_type  GOP = _params->GOP, GEP = _params->GEP, w_max = 0;
	sw_link		*sw_links;
	sw_cell		*current_line = G_sw_compute_data.line1, *prev_line = G_sw_compute_data.line2, *p = NULL;
	//int*		path = G_sw_compute_data.path;
	char		dir, open_flag;

	assert(_seq1);
	assert(_seq2);
	assert(_params);

	// initialization
	G_sw_compute_data.seq1 = _seq1;
	G_sw_compute_data.seq2 = _seq2;

	// general memory allocation for NW-graph
	sw_links = (sw_link*)malloc(m * n * sizeof(sw_link));
	assert(sw_links);
	memset(sw_links, 0, m * n * sizeof(sw_link));

	// prepare
	memset(current_line, 0, (n + 1) * sizeof(sw_cell));
	memset(prev_line, 0, (n + 1) * sizeof(sw_cell));
	current_line++;
	prev_line++;

	//   
	for(i = 0; i < m; i++)
	{
		// swap prev/current
		p = prev_line;
		prev_line = current_line;
		current_line = p;

		//  
		for(j = 0; j < n; j++) {

			current_line[j].diag = prev_line[j-1].diag + subst_matrix[seq1[i]][seq2[j]];

			// 5-comparison goes here
			dir = 0; open_flag = 0;

			if(current_line[j-1].left < current_line[j-1].diag - GOP) {
				open_flag |= DIR_LEFTOPEN;
				current_line[j].left = current_line[j-1].diag - GOP - GEP;
			} else 
				current_line[j].left = current_line[j-1].left - GEP;

			if(prev_line[j].top < prev_line[j].diag - GOP) {
				open_flag |= DIR_TOPOPEN;
				current_line[j].top = prev_line[j].diag - GOP - GEP;
			} else 
				current_line[j].top = prev_line[j].top - GEP;

			if(current_line[j].top > current_line[j].diag) {
				if(current_line[j].top > current_line[j].left) {
					current_line[j].diag = current_line[j].top;
					dir = DIR_TOP;
				} else {
					current_line[j].diag = current_line[j].left;
					dir = DIR_LEFT;
				}
			} else if(current_line[j].left > current_line[j].diag) {
				current_line[j].diag = current_line[j].left;
				dir = DIR_LEFT;
			} else {
				dir = DIR_DIAG;
			}

			if(current_line[j].diag < 0) {
				dir = DIR_TERMINATOR;
				current_line[j].diag = 0;
			} else if(current_line[j].diag >= w_max) {
				w_max = current_line[j].diag;
				i_max = i;
				j_max = j;
			}

			sw_links[i * n + j].link = dir;
			sw_links[i * n + j].open_flag = open_flag;
		}
	}

	free(sw_links);

	G_sw_compute_data.max_weight = w_max;

	return SUCCESS;
}

//------------------------------------------------------------------------------
int compute_sw_alignment_cut_seqsubst(	// need activated seq_matrix
	sequence*	_seq1,					// with _seq1
	sequence*	_seq2,
	sw_params*	_params
)
{
	unsigned	n = _seq1->length, m = _seq2->length;
	int			i, j, i_max = 0, j_max = 0;
	char		*seq0 = _seq1->aas, *seq = _seq2->aas;
	score_type  GOP = _params->GOP, GEP = _params->GEP, w_max = 0;
	sw_link		*sw_links;
	sw_cell		*current_line = G_sw_compute_data.line1, *prev_line = G_sw_compute_data.line2, *p = NULL;
	//int*		path = G_sw_compute_data.path;
	char		dir, open_flag;
	score_type  *letter_subst;

	assert(_seq1);
	assert(_seq2);
	assert(_params);

	// initialization
	//G_sw_compute_data.seq1 = _seq1;
	//G_sw_compute_data.seq2 = _seq2;

	// general memory allocation for NW-graph
	sw_links = (sw_link*)malloc(m * n * sizeof(sw_link));
	assert(sw_links);
	memset(sw_links, 0, m * n * sizeof(sw_link));

	// prepare
	memset(current_line, 0, (n + 1) * sizeof(sw_cell));
	memset(prev_line, 0, (n + 1) * sizeof(sw_cell));
	current_line++;
	prev_line++;

	//   
	for(i = 0; i < m; i++)
	{
		// swap prev/current
		p = prev_line;
		prev_line = current_line;
		current_line = p;

		letter_subst = seq_matrix + (seq[i] * n);

		//  
		for(j = 0; j < n; j++) {

			current_line[j].diag = prev_line[j-1].diag + (*letter_subst); //subst_matrix[seq1[i]][seq2[j]];
			letter_subst ++;

			// 5-comparison goes here
			dir = 0; open_flag = 0;

			if(current_line[j-1].left < current_line[j-1].diag - GOP) {
				open_flag |= DIR_LEFTOPEN;
				current_line[j].left = current_line[j-1].diag - GOP - GEP;
			} else 
				current_line[j].left = current_line[j-1].left - GEP;

			if(prev_line[j].top < prev_line[j].diag - GOP) {
				open_flag |= DIR_TOPOPEN;
				current_line[j].top = prev_line[j].diag - GOP - GEP;
			} else 
				current_line[j].top = prev_line[j].top - GEP;

			if(current_line[j].top > current_line[j].diag) {
				if(current_line[j].top > current_line[j].left) {
					current_line[j].diag = current_line[j].top;
					dir = DIR_TOP;
				} else {
					current_line[j].diag = current_line[j].left;
					dir = DIR_LEFT;
				}
			} else if(current_line[j].left > current_line[j].diag) {
				current_line[j].diag = current_line[j].left;
				dir = DIR_LEFT;
			} else {
				dir = DIR_DIAG;
			}

			if(current_line[j].diag < 0) {
				dir = DIR_TERMINATOR;
				current_line[j].diag = 0;
			} else if(current_line[j].diag >= w_max) {
				w_max = current_line[j].diag;
				i_max = i;
				j_max = j;
			}

			sw_links[i * n + j].link = dir;
			sw_links[i * n + j].open_flag = open_flag;
		}
	}

	free(sw_links);
	G_sw_compute_data.max_weight = w_max;
	return SUCCESS;
}

//--------------------------------------------------------------------------------
int compute_sw_alignment_dblgap(
	sequence*	_seq1,
	sequence*	_seq2,
	sw_params*	_params
)
{
	unsigned	m = _seq1->length, n = _seq2->length;
	int			i, j, i_max = 0, j_max = 0;
	char		*seq1 = _seq1->aas, *seq2 = _seq2->aas;
	score_type	GOP = _params->GOP, GEP = _params->GEP, w_max = 0, jump_left = 0, jump_top = 0;
	sw_link		*sw_links;
	sw_cell		*current_line = G_sw_compute_data.line1, *prev_line = G_sw_compute_data.line2, *p = NULL;
	int*		path = G_sw_compute_data.path;
	char		dir, open_flag;

	assert(_seq1);
	assert(_seq2);
	assert(_params);

	// initialization
	G_sw_compute_data.seq1 = _seq1;
	G_sw_compute_data.seq2 = _seq2;

	// general memory allocation for NW-graph
	sw_links = (sw_link*)malloc(m * n * sizeof(sw_link));
	assert(sw_links);
	memset(sw_links, 0, m * n * sizeof(sw_link));

	// prepare
	memset(current_line, 0, (m + 1) * sizeof(sw_cell));
	memset(prev_line, 0, (m + 1) * sizeof(sw_cell));
	current_line++;
	prev_line++;

	//   
	for(j = 0; j < n; j++)
	{
		//  
		for(i = 0; i < m; i++) {

			current_line[i].diag = prev_line[i-1].diag + subst_matrix[seq1[i]][seq2[j]];

			// 5-comparison goes here
			dir = 0; open_flag = 0;

			if(current_line[i-1].jump < current_line[i-1].diag - GOP) {
				open_flag |= DIR_LEFTOPEN;
				jump_left = current_line[i-1].diag - GOP - GEP;
			} else 
				jump_left = current_line[i-1].jump - GEP;

			if(prev_line[i].jump < prev_line[i].diag - GOP) {
				open_flag |= DIR_TOPOPEN;
				jump_top = prev_line[i].diag - GOP - GEP;
			} else 
				jump_top = prev_line[i].jump - GEP;

			if(jump_top > jump_left) {
				current_line[i].jump = jump_top;
				dir = DIR_TOP;
			} else {
				current_line[i].jump = jump_left;
				dir = DIR_LEFT;
			}

			if(current_line[i].jump > current_line[i].diag) {
				current_line[i].diag = current_line[i].jump;
			} else {
				dir = DIR_DIAG;
			}

			if(current_line[i].diag < 0) {
				current_line[i].diag = 0;
				dir = DIR_TERMINATOR;
			}


			if(current_line[i].diag >= w_max) {
				w_max = current_line[i].diag;
				i_max = i; j_max = j;
			}

			sw_links[j * m + i].link = dir;
			sw_links[j * m + i].open_flag = open_flag;
		}
		// swap prev/current
		p = prev_line;
		prev_line = current_line;
		current_line = p;
	}

	//  ,    

	for(i = m - 1; i > i_max; path[i--] = -1);
	i = i_max; j = j_max;
	while(i >= 0 && j >= 0) {
		switch(sw_links[j * m + i].link) {
		case DIR_DIAG:
			path[i--] = j--;
			break;
		case DIR_TOP :
			do {
				if(sw_links[j * m + i].link == DIR_LEFT) break;
				assert(j >= 0);
				path[i] = j; 
			} while (!(sw_links[(j--) * m + i].open_flag & DIR_TOPOPEN));
			break;
		case DIR_LEFT :
			do {
				if(sw_links[j * m + i].link == DIR_TOP) break;
				assert(i >= 0);
				path[i] = -1;
			} while (!(sw_links[j * m + i--].open_flag & DIR_LEFTOPEN));
			break;
		default:
			j = -1; // in order to exit while !!!
			break;
		}
	}

	while(i >= 0) path[i--] = -1;

	G_sw_compute_data.max_weight = w_max;

	free(sw_links);
	//free(--current_line);
	//free(--prev_line);

	return SUCCESS;
}

//------------------------------------------------------------
void sw_data_release(sw_compute_data* _data)
{
	// free(_data->path);
	// _data->path = NULL;
};

//------------------------------------------------------------
int format_sw_alignment_output(sw_compute_data* _data, char* _align1, char* _align2)
{
	int i, j, m = _data->seq1->length, i_pos = 0, j_pos = 0;
	int* path = _data->path;

	assert(_data);
	assert(_align1);
	assert(_align2);

	for(i = 0; i < m; i++) {
		j = path[i];
		// reach the (i, j)
		if(j == -1) {
			*_align1++ = amino_acids[_data->seq1->aas[i]];
			*_align2++ = GAP_SYMBOL;
		} else {
			while(j_pos < j)  {   // reach the j
				*_align1++ = GAP_SYMBOL;
				*_align2++ = amino_acids[_data->seq2->aas[j_pos++]];
			}
			*_align1++ = amino_acids[_data->seq1->aas[i]];
			*_align2++ = amino_acids[_data->seq2->aas[j_pos++]];
		}
	}
	return SUCCESS;
}